# Copyright 2011-2017 Kwant authors.
#
# This file is a modified version of a file that is part of Kwant.
# It is subject to the license terms below. A list of Kwant authors
# can be found at http://kwant-project.org/authors.
#
# =============
# Kwant license
# =============
#
# Copyright 2011-2015 C. W. Groth (CEA), M. Wimmer, A. R. Akhmerov,
# X. Waintal (CEA), and others.  All rights reserved.
#
# (CEA = Commissariat à l'énergie atomique et aux énergies alternatives)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import keyword
from collections import defaultdict

import numpy as np

import sympy
import sympy.abc
import sympy.physics.quantum
from sympy.core.function import AppliedUndef
from sympy.core.sympify import converter
from sympy.core.core import all_classes as sympy_classes
from sympy.physics.matrices import msigma as _msigma

import warnings

# TODO: remove when sympy correctly includes MutableDenseMatrix (lol).
sympy_classes = set(sympy_classes) | {sympy.MutableDenseMatrix}

momentum_operators = sympy.symbols('k_x k_y k_z', commutative=False)
position_operators = sympy.symbols('x y z', commutative=False)

pauli = [sympy.eye(2), _msigma(1), _msigma(2), _msigma(3)]

extra_ns = sympy.abc._clash.copy()
extra_ns.update({s.name: s for s in momentum_operators})
extra_ns.update({s.name: s for s in position_operators})
extra_ns.update({'kron': sympy.physics.quantum.TensorProduct,
                 'eye': sympy.eye, 'identity': sympy.eye})
extra_ns.update({'sigma_{}'.format(c): p for c, p in zip('0xyz', pauli)})


# workaroud for https://github.com/sympy/sympy/issues/12060
del extra_ns['I']
del extra_ns['pi']


################  Helpers to handle sympy

def lambdify(expr, locals=None):
    """Return a callable object for computing a continuum Hamiltonian.

    .. warning::
        This function uses ``eval`` (because it calls ``sympy.sympify``), and
        thus should not be used on unsanitized input.

    If necessary, the given expression is sympified using
    `kwant.continuum.sympify`.  It is then converted into a callable object.

    Parameters
    ----------
    expr : str or SymPy expression
        Expression to be converted into a callable object
    locals : dict or ``None`` (default)
        Additional definitions for `~kwant.continuum.sympify`.

    Examples
    --------
    >>> f = lambdify('a + b', locals={'b': 'b + c'})
    >>> f(1, 3, 5)
    9

    >>> ns = {'sigma_plus': [[0, 2], [0, 0]]}
    >>> f = lambdify('k_x**2 * sigma_plus', ns)
    >>> f(0.25)
    array([[ 0.   ,  0.125],
           [ 0.   ,  0.   ]])
    """
    expr = sympify(expr, locals)

    args = [s.name for s in expr.atoms(sympy.Symbol)]
    args += [str(f.func) for f in expr.atoms(AppliedUndef, sympy.Function)]

    return sympy.lambdify(sorted(args), expr)


def sympify(expr, locals=None):
    """Sympify object using special rules for Hamiltonians.

    If `'expr`` is already a type that SymPy understands, it will do nothing
    but return that value. Note that ``locals`` will not be used in this
    situation.

    Otherwise, it is sympified by ``sympy.sympify`` with a modified namespace
    such that

    * the position operators "x", "y" or "z" and momentum operators "k_x",
      "k_y", and "k_z" do not commute,
    * all single-letter identifiers and names of Greek letters (e.g. "pi" or
      "gamma") are treated as symbols,
    * "kron" corresponds to ``sympy.physics.quantum.TensorProduct``, and
      "identity" to ``sympy.eye``,
    * "sigma_0", "sigma_x", "sigma_y", "sigma_z" are the Pauli matrices.

    In addition, Python list literals are interpreted as SymPy matrices.

    .. warning::
        This function uses ``eval`` (because it calls ``sympy.sympify``), and
        thus should not be used on unsanitized input.

    Parameters
    ----------
    expr : str or SymPy expression
        Expression to be converted to a SymPy object.
    locals : dict or ``None`` (default)
        Additional entries for the namespace under which `expr` is sympified.
        The keys must be valid Python variable names.  The values may be
        strings, since they are all are sent through `continuum.sympify`
        themselves before use.  (Note that this is a difference to how
        ``sympy.sympify`` behaves.)

        .. note::
            When a value of `locals` is already a SymPy object, it is used
            as-is, and the caller is responsible to set the commutativity of
            its symbols appropriately.  This possible source of errors is
            demonstrated in the last example below.

    Returns
    -------
    result : SymPy object

    Examples
    --------
        >>> sympify('k_x * A(x) * k_x + V(x)')
        k_x*A(x)*k_x + V(x)     # as valid sympy object

        >>> sympify('k_x**2 + V', locals={'V': 'V_0 + V(x)'})
        k_x**2 + V(x) + V_0

        >>> ns = {'sigma_plus': [[0, 2], [0, 0]]}
        >>> sympify('k_x**2 * sigma_plus', ns)
        Matrix([
        [0, 2*k_x**2],
        [0,        0]])

        >>> sympify('k_x * A(c) * k_x', locals={'c': 'x'})
        k_x*A(x)*k_x
        >>> sympify('k_x * A(c) * k_x', locals={'c': sympy.Symbol('x')})
        A(x)*k_x**2

    """
    stored_value = None

    # if ``expr`` is already a ``sympy`` object we may terminate a code path
    if isinstance(expr, tuple(sympy_classes)):
        if locals:
            warnings.warn('Input expression is already SymPy object: '
                          '"locals" will not be used.',
                          RuntimeWarning,
                          stacklevel=2)

        # We assume that all present functions, like "sin", "cos", will be
        # provided by user during the final evaluation through "params".
        # Therefore we make sure they are defined as AppliedUndef, not built-in
        # sympy types.
        subs = {r: sympy.Function(str(r.func))(*r.args)
                for r in expr.atoms(sympy.Function)}

        return expr.subs(subs)

    # if ``expr`` is not a "sympy" then we proceed with sympifying process
    if locals is None:
        locals = {}

    for k in locals:
        if (not isinstance(k, str)
            or not k.isidentifier() or keyword.iskeyword(k)):
            raise ValueError(
                "Invalid key in 'locals': {}\nKeys must be "
                "identifiers and may not be keywords".format(repr(k)))

    # sympify values of locals before updating it with extra_ns
    locals = {k: sympify(v) for k, v in locals.items()}
    for k, v in extra_ns.items():
        locals.setdefault(k, v)
    try:
        stored_value = converter.pop(list, None)
        converter[list] = lambda x: sympy.Matrix(x)
        hamiltonian = sympy.sympify(expr, locals=locals)
        # if input is for example ``[[k_x * A(x) * k_x]]`` after the first
        # sympify we are getting list of sympy objects, so we call sympify
        # second time to obtain ``sympy`` matrices.
        hamiltonian = sympy.sympify(hamiltonian)
    finally:
        if stored_value is not None:
            converter[list] = stored_value
        else:
            del converter[list]

    return hamiltonian


def make_commutative(expr, *symbols):
    """Make sure that specified symbols are defined as commutative.

    Parameters
    ----------
    expr: sympy.Expr or sympy.Matrix
    symbols: sequace of symbols
        Set of symbols that are requiered to be commutative. It doesn't matter
        of symbol is provided as commutative or not.

    Returns
    -------
    input expression with all specified symbols changed to commutative.
    """
    symbols = [sympy.Symbol(s.name, commutative=False) for s in symbols]
    expr = expr.subs({s: sympy.Symbol(s.name) for s in symbols})
    return expr


def monomials(expr, gens=None):
    """Parse ``expr`` into monomials in the symbols in ``gens``.

    Parameters
    ----------
    expr: sympy.Expr or sympy.Matrix
        Sympy expression to be parsed into monomials.
    gens: sequence of sympy.Symbol objects or strings (optional)
        Generators of monomials. If unset it will default to all
        symbols used in ``expr``.

    Returns
    -------
    dictionary (generator: monomial)

    Examples
    --------
        >>> expr = kwant.continuum.sympify("A * (x**2 + y) + B * x + C")
        >>> monomials(expr, gens=('x', 'y'))
        {1: C, x: B, x**2: A, y: A}
    """
    if gens is None:
        gens = expr.atoms(sympy.Symbol)
    else:
        gens = [sympify(g) for g in gens]

    if not isinstance(expr, sympy.MatrixBase):
        return _expression_monomials(expr, gens)
    else:
        output = defaultdict(lambda: sympy.zeros(*expr.shape))
        for (i, j), e in np.ndenumerate(expr):
            mons = _expression_monomials(e, gens)
            for key, val in mons.items():
                output[key][i, j] += val
        return dict(output)


def _expression_monomials(expr, gens):
    """Parse ``expr`` into monomials in the symbols in ``gens``.

    Parameters
    ----------
    expr: sympy.Expr
        Sympy expr to be parsed.
    gens: sequence of sympy.Symbol
        Generators of monomials.

    Returns
    -------
    dictionary (generator: monomial)
    """
    expr = sympy.expand(expr)
    output = defaultdict(lambda: sympy.Integer(0))
    for summand in expr.as_ordered_terms():
        key = []
        val = []
        for factor in summand.as_ordered_factors():
            symbol, exponent = factor.as_base_exp()
            if symbol in gens:
                key.append(factor)
            else:
                val.append(factor)
        output[sympy.Mul(*key)] += sympy.Mul(*val)

    return dict(output)


################ general help functions

def gcd(*args):
    if len(args) == 1:
        return args[0]

    L = list(args)

    while len(L) > 1:
        a = L[len(L) - 2]
        b = L[len(L) - 1]
        L = L[:len(L) - 2]

        while a:
            a, b = b%a, a

        L.append(b)

    return abs(b)



import pytest
import sympy
import numpy as np
import scipy.linalg as la

from .. import kwant_rmt
from ..hamiltonian_generator import continuum_hamiltonian, check_symmetry, \
     bloch_family, make_basis_pretty, constrain_family, continuum_variables, \
     continuum_pairing, remove_duplicates, subtract_family
from ..groups import PointGroupElement, Model, time_reversal, chiral, rotation
from ..model import _commutative_momenta, Model, BlochModel
from ..linalg import nullspace, family_to_vectors


def test_check_symmetry():
    """Test discrete symmetries for the Hamiltonian generator without momentum
    dependence, and that antiunitary and anticommuting symmetries are correctly
    treated in check_symmetry. """

    # No momentum dependence
    dim = 0
    total_power = 0
    n = 8
    R = np.eye(dim, dtype=int)

    for sym in kwant_rmt.sym_list:
        symmetries = []
        if kwant_rmt.p(sym):
            p_mat = np.array(kwant_rmt.h_p_matrix[sym])
            p_mat = np.kron(np.identity(n // len(p_mat)), p_mat)
            symmetries.append(PointGroupElement(R, True, True, p_mat))
        if kwant_rmt.t(sym):
            t_mat = np.array(kwant_rmt.h_t_matrix[sym])
            t_mat = np.kron(np.identity(n // len(t_mat)), t_mat)
            symmetries.append(PointGroupElement(R, True, False, t_mat))
        if kwant_rmt.c(sym):
            c_mat = np.kron(np.identity(n // 2), np.diag([1, -1]))
            symmetries.append(PointGroupElement(R, False, True, c_mat))

        if len(symmetries):
            Hamiltonian_family = continuum_hamiltonian(symmetries, dim, total_power)
            # Check that the symmetries hold.
            check_symmetry(Hamiltonian_family, symmetries)
            # Also manually check that the family has the symmetries, to test
            # the function check_symmetry.
            # Check several random linear combinations of all family members.
            for _ in range(5):
                ham = sum([np.random.rand()*matrix for member in Hamiltonian_family
                           for matrix in member.values()])
                # Iterate over all symmetries
                for symmetry in symmetries:
                    Left = ham.dot(symmetry.U)
                    if symmetry.conjugate: # Symmetry is antiunitary
                        Right = symmetry.U.dot(ham.conj())
                    else:  # Symmetry is unitary
                        Right = symmetry.U.dot(ham)
                    if symmetry.antisymmetry: # Symmetry anticommutes
                        assert np.allclose(Left + Right, 0)
                    else: # Symmetry commutes
                        assert np.allclose(Left - Right, 0)


def test_bloch_generator():
    """Square lattice with time reversal and rotation symmetry, such that all hoppings are real. """
    # Time reversal
    trU = np.eye(2)
    trR = sympy.Matrix(np.eye(2, dtype=int))
    trS = PointGroupElement(trR, True, False, trU)
    # 2-fold rotation
    rotU = np.eye(2)
    sphi = sympy.pi
    rotR = sympy.Matrix([[sympy.cos(sphi), -sympy.sin(sphi)],
                         [sympy.sin(sphi), sympy.cos(sphi)]])
    rotS = PointGroupElement(rotR, False, False, rotU)
    symmetries = [rotS, trS]

    # With integer hopping, output is list of Model
    hopping_vectors = [(0, 0, np.array([0, 1])), (0, 0, np.array([1, 0]))]
    norbs = [(0, 2)]
    family = bloch_family(hopping_vectors, symmetries, norbs, onsites=False)
    assert len(family) == 6, 'Incorrect number of members in the family'
    check_symmetry(family, symmetries)
    pretty = make_basis_pretty(family, num_digits=3)
    check_symmetry(pretty, symmetries)
    # All members should be real given the constraints.
    assert all([np.allclose(value, value.conj()) for member in pretty for value in member.values()])
    assert all([np.allclose(value, value.conj()) for member in family for value in member.values()])
    # Constraining the family again should have no effect
    again = constrain_family(symmetries, family)
    check_symmetry(again, symmetries)
    for member in family:
        assert any([member == other for other in again])

    # With floating point hopping, output is list of BlochModel
    hopping_vectors = [(0, 0, np.array([0, 0.5])), (0, 0, np.array([0.5, 0]))]
    norbs = [(0, 2)]
    family = bloch_family(hopping_vectors, symmetries, norbs, onsites=False, bloch_model=True)
    assert len(family) == 6, 'Incorrect number of members in the family'
    check_symmetry(family, symmetries)
    pretty = make_basis_pretty(family, num_digits=3)
    check_symmetry(pretty, symmetries)
    # All members should be real given the constraints.
    assert all([np.allclose(value, value.conj()) for member in pretty for value in member.values()])
    assert all([np.allclose(value, value.conj()) for member in family for value in member.values()])
    # Constraining the family again should have no effect
    again = constrain_family(symmetries, family)
    check_symmetry(again, symmetries)
    for member in family:
        assert any([member == other for other in again])


def test_graphene():
    # Test Model
    # Time reversal
    TR = time_reversal(2, U=np.eye(2))
    # Chiral symmetry
    C = chiral(2, U=np.array([[1, 0], [0, -1]]))
    # Atom A rotates into A, B into B, use exact sympy representation
    sphi = 2*sympy.pi/3
    RC3 = sympy.Matrix([[sympy.cos(sphi), -sympy.sin(sphi)],
                      [sympy.sin(sphi), sympy.cos(sphi)]])
    C3 = PointGroupElement(RC3, False, False, np.eye(2))
    symmetries = [C, TR, C3]
    norbs = [('A', 1), ('B', 1)]  # A and B atom per unit cell, one orbital each
    hopping_vectors = [('A', 'B', [0, 1])] # Hopping between neighbouring A and B atoms
    family = bloch_family(hopping_vectors, symmetries, norbs)
    assert len(family) == 1
    result = Model({'e**(sqrt(3)*I*k_x/2 + I*k_y/2)': np.array([[0, 0], [1, 0]]),
                    'e**(-sqrt(3)*I*k_x/2 - I*k_y/2)': np.array([[0, 1], [0, 0]]),
                    'e**(sqrt(3)*I*k_x/2 - I*k_y/2)': np.array([[0, 1], [0, 0]]),
                    'e**(-I*k_y)': np.array([[0, 0], [1, 0]]),
                    'e**(-sqrt(3)*I*k_x/2 + I*k_y/2)': np.array([[0, 0], [1, 0]]),
                    'e**(I*k_y)': np.array([[0, 1], [0, 0]])
                    },
                   normalize=True,
                   momenta=('k_x', 'k_y'))
    assert family[0] == result

    # Test BlochModel
    C3 = rotation(1/3, U=np.eye(2))
    symmetries = [C, TR, C3]
    family = bloch_family(hopping_vectors, symmetries, norbs, bloch_model=True)
    assert len(family) == 1
    assert family[0] == BlochModel(result)


def test_continuum_variables():
    dim = 2
    total_power = 2

    # Test default arguments
    momenta = _commutative_momenta
    terms = continuum_variables(dim, total_power)
    result = [momenta[0]**0, momenta[0], momenta[1], momenta[0]**2, momenta[1]**2, momenta[1]*momenta[0]]
    assert all([term in result for term in terms]), 'Incorrect momentum terms returned by continuum_variables.'

    # Limit some powers of k
    all_powers = [1, 2]
    terms = continuum_variables(dim, total_power, all_powers=all_powers)
    result = [momenta[0]**0, momenta[0], momenta[1], momenta[1]**2, momenta[1]*momenta[0]]
    assert all([term in result for term in terms]), 'Incorrect limitations on powers in continuum_variables.'
    all_powers = [[1], [1, 3]]
    momenta = _commutative_momenta
    result = [momenta[1]*momenta[0]]
    terms = continuum_variables(dim, total_power, all_powers=all_powers)
    assert all([term in result for term in terms]), 'Incorrect limitations on powers in continuum_variables.'

    # Test different momenta
    dim = 2
    total_power = 4
    all_powers = [[1], [1, 3]]
    momenta = [sympy.Symbol('k_1'), sympy.Symbol('k_a')]
    result = [momenta[1]*momenta[0], momenta[0]*momenta[1]**3]
    terms = continuum_variables(dim, total_power, all_powers=all_powers, momenta=momenta)
    assert all([term in result for term in terms]), 'New momentum variables treated incorrectly.'


def test_pairing_generator():

    mU = np.array([[0.0, 1.0j],
                   [1.0j, 0.0]])
    mR = np.array([[-1]])

    # Use the pairing generator
    ph_squares = [1, -1]
    phases = np.hstack((np.exp(1j*np.random.rand(1)), np.array([1])))
    for ph_square in ph_squares:
        for phase in phases:
            mS = PointGroupElement(mR, False, False, mU)
            symmetries = [mS]
            dim = 1  # Momenta along x
            total_power = 1 # Maximum power of momenta
            from_pairing = continuum_pairing(symmetries, dim, total_power, ph_square=ph_square,
                                             phases=(phase, ), prettify=False)
            # Use the continuum Hamiltonian generator
            # Mirror
            mS = PointGroupElement(mR, False, False, mU)
            symmetries = [mS]
            # Extend the symmetry to particle-hole space
            N = mS.U.shape[0]
            mS.U = la.block_diag(mS.U, phase*mS.U.conj())
            # Build ph operator
            ph = np.array([[0, 1], [ph_square, 0]])
            ph = PointGroupElement(np.eye(dim), True, True, np.kron(ph, np.eye(N)))
            symmetries.append(ph)
            # Pick out pairing blocks
            cont_ham = continuum_hamiltonian(symmetries, dim, total_power)
            from_ham = [Model({term: matrix[:N, N:] for term, matrix in monomial.items()}) for monomial in
                          cont_ham]
            from_ham = [mon for mon in from_ham if len(mon)]
            from_ham = remove_duplicates(from_ham)

            assert len(from_pairing) == len(from_ham)
            # Combine into one redundant family, and then remove the redundancies.
            remove_dupls = remove_duplicates(from_pairing + from_ham)
            assert len(remove_dupls) == len(from_ham)


def test_subtract_family():
    paulis = [np.eye(2),
              np.array([[0, 1], [1, 0]]),
              np.array([[0, -1j], [1j, 0]]),
              np.array([[1, 0], [0, -1]])]
    One = _commutative_momenta[0]**0
    fam1 = [Model({One: pmat*(0.5 + np.random.rand())}) for pmat in paulis]
    fam2 = [Model({One: paulis[2] + paulis[3]}), Model({One: paulis[2] - paulis[3]})]
    result = subtract_family(fam1, fam2)
    correct = [Model({One: pmat}) for pmat in paulis[:2]]  # Correct span
    # Check that the correct result and the returned result have the same span
    null_mat = np.hstack((family_to_vectors(result).T, -family_to_vectors(correct).T))
    null_basis = nullspace(null_mat)
    assert null_basis.shape[0] == len(correct)


def test_pretty_basis():
    sx = np.array([[0, 1], [1, 0]])
    sz = np.array([[1, 0], [0, -1]])
    One = _commutative_momenta[0]**0
    fam = [Model({One: sx*(0.5 + np.random.rand()) + sz*(0.5 + np.random.rand())}),
           Model({One: sx*(0.5 + np.random.rand()) - sz*(0.5 + np.random.rand())})]
    pfam = make_basis_pretty(fam, num_digits=3)
    assert len(pfam) == len(fam), 'Basis sparsification failed, returned a smaller basis.'
    # The sparse basis should be sx and sz
    for member in pfam:
        mat = list(member.values())[0]
        if np.isclose(mat[0, 0], 0):
            assert np.allclose(mat/mat[0, 1], sx), 'Sparsification failed.'
        else:
            assert np.allclose(mat/mat[0, 0], sz), 'Sparsification failed.'




import warnings
import numpy as np
import scipy.linalg as la
import scipy.sparse.linalg as sla
import scipy
from numbers import Number
from scipy.optimize import minimize
from scipy.spatial.distance import cdist
from scipy.sparse.csgraph import connected_components
import itertools as it
import tinyarray as ta


def commutator(A, B):
    return np.dot(A, B) - np.dot(B, A)


def prop_to_id(A):
    # Test if A is proportional to the identity matrix
    # and return the factor as well
    if np.isclose(A[0, 0], 0):
        if np.allclose(A, np.zeros(A.shape)):
            return True, 0
        else:
            return False, 0
    else:
        if np.allclose(A / A[0, 0], np.eye(*A.shape)):
            return (True, A[0, 0])
        else:
            return False, 0


def allclose(a, b, rtol=1e-05, atol=1e-08, equal_nan=False):
    """Alternative to numpy.allclose to test that two ndarrays
    are elementwise close. Unlike the numpy version, the relative
    tolerance is not elementwise, but it is relative to
    the largest entry in the arrays."""
    # If either is a list or number, recast it to ndarray, if it is
    # a LinearOperator, recast it to dense matrix. This is rather
    # inefficient, but LinearOperator doesn't support multiplication
    # with sparse matrices.
    if isinstance(a, (list, Number)):
        a = np.asarray(a)
    elif isinstance(a, sla.LinearOperator):
        a = a @ np.eye(a.shape[-1], dtype=complex)
    if isinstance(b, (list, Number)):
        b = np.asarray(b)
    elif isinstance(b, sla.LinearOperator):
        b = b @ np.eye(a.shape[-1], dtype=complex)

    # Check if empty arrays, only compare shape
    if a.size == 0:
        return a.shape == b.shape
    atol = atol + rtol * max(np.max(np.abs(a)), np.max(np.abs(b)))
    diff = a - b
    if isinstance(diff, scipy.sparse.spmatrix):
        diff = diff.data

    return np.allclose(diff, 0, rtol=0, atol=atol, equal_nan=equal_nan)


def mtm(a, B, c):
    # matrix-tensor-matrix multiplication for dimensions 2, 3, 2.
    # Equivalent to 'np.einsum('ij,ajk,kl->ail', a, B, c)'
    # and to 'a @ B @ c', but much faster.
    return np.array([a.dot(b).dot(c) for b in B])


def matrix_basis(dim, traceless=False, antihermitian=False, real=False, sparse=False):
    """"Construct a basis for the vector space of dim x dim matrices,
    that may in addition be traceless, and/or composed of Hermitian
    or skew-Hermitian basis elements.

    Parameters
    ----------
    dim: positive integer
        The dimension of the matrices (dim x dim).
    traceless: boolean
        Whether the vector space only includes traceless matrices (True)
        or not (False).
    antihermitian: boolean
        Whether to use a basis composed of Hermitian matrices (False),
        or skew-Hermitian matrices (True).
    real:
        Whether the basis only spans real matrices. Depending on
        'antihermitian' it spans symmetric (False) or antisymmetric
        (True) matrices only.
    sparse:
        If True, scipy.sparse.csr_matrix is returned, otherwise
        np.ndarray.

    Returns
    ---------
    basis: generator
        A generator that returns matrices that span the vector space.
    """
    if sparse:
        null = lambda: scipy.sparse.lil_matrix((dim, dim), dtype=complex)
        set_type = lambda x: x.tocsr()
    else:
        null = lambda: np.zeros((dim, dim), dtype=complex)
        set_type = lambda x: x
    # Matrix basis for dim x dim matrices. With real coefficients spans
    # Hermitian matrices, with complex spans all matrices
    coeff = (1j if antihermitian else 1)
    # Diagonals
    if real and antihermitian:
        pass
    elif traceless:
        for i in range(dim - 1):
            diag = null()
            diag[i, i] = 1
            diag[dim-1, dim-1] = -1
            yield coeff * set_type(diag)
    else:
        for i in range(dim):
            diag = null()
            diag[i, i] = 1
            yield coeff * set_type(diag)
    # Off-diagonals
    for i, j in it.combinations(list(range(dim)), 2):
        if not (real and antihermitian):
            h = null()
            h[i, j] = 1
            h[j, i] = 1
            yield coeff * set_type(h)
        if not (real and not antihermitian):
            h = null()
            h[i, j] = 1j
            h[j, i] = -1j
            yield coeff * set_type(h)


def family_to_vectors(family, all_keys=None):
    """Convert a list of Model to standard vector representation.

    Parameters
    ----------
    family: iterable of Model objects
        The Model to convert to vector representation.
    all_keys: iterable or None
        Iterable that must contain all the keys in the family.
        An optional parameter that allows to fix the ordering of
        the returned vectors to the same as in all_keys.
        Otherwise, the keys are ordered arbitrarily (all_keys = None).

    Returns
    ----------
    A matrix representing the family, in which the family members form the
    rows."""

    def matrix_to_vector(M):
        M = M.reshape(-1)
        return np.concatenate((M.real, M.imag))

    if all_keys is None:
        all_keys = set()
        for term in family:
            all_keys |= term.keys()
        all_keys = list(all_keys)
    assert all([set(all_keys) >= term.keys() for term in family])
    if not all_keys:
        return np.empty((len(family), 0))
    # Map each family member to a vector, gather the vectors into
    # a matrix in which the vectors form the rows.
    basis_vectors = []
    for member in family:
        vector = [matrix_to_vector(mat) for mat in member.value_list(all_keys)]
        basis_vectors.append(np.hstack(vector))
    return np.vstack(basis_vectors)


def nullspace(A, atol=1e-6, return_complement=False, sparse=None, k_max=-10):
    """Compute an approximate basis for the nullspace of matrix A.

    Parameters:
    -----------
    A : ndarray or spmatrix
        A should be 2-D.
    atol : real
        Absolute tolerance when deciding whether an eigen/singular
        value is zero, so the corresponding vector is included in the
        null space.
    return_complement : bool
        Whether to return the basis of the orthocomplement ot the
        null space as well. Treated as False if sparse is True.
    sparse : bool  or None (default)
        Whether to use sparse linear algebra to find the null space.
        The default value is inferred from the type of the input array.
    k_max : int (default -10)
        If k_max<0, the full null-space is found by increasing the number
        of eigenvectors found in each step by abs(k_max).
        If k_max>0, k_max is the maximum number of null space vectors requested.
        If the null space dimension is higher than k_max, basis of a random
        subspace is returned. Can be used to increase performance if the
        maximum null space dimensionality is known.
        Ignored unless sparse is True.

    Returns:
    --------
    ns : ndarray
        Basis of the null space of A as row vectors.
    nsc : ndarray
        Basis of the complement of the null space as row vectors, the
        othonormal basis of the row span of A. Only returned if
        return_complement=True and sparse=False.
    """
    if sparse is None:
        sparse = isinstance(A, scipy.sparse.spmatrix)
    if sparse:
        # Do sparse eigenvalue solving
        # sigma used for shift-invert, should be a small negative value
        sigma = -1e-1
        # Make sure A is sparse matrix
        if isinstance(A, np.ndarray):
            A = scipy.sparse.csr_matrix(A)
        A.eliminate_zeros()
        # Treat A=0 case
        if np.allclose(A.data, 0, atol=atol/A.shape[1]):
            return np.eye(A.shape[1])
        # Make Hermitian positive definite matrix
        A = A.T.conj().dot(A)

        if k_max > 0:
            # If k_max is specified, find k_max smallest eigenvalues
            evals, evecs = sla.eigsh(A, sigma=sigma, which='LM',
                                     k=k_max, return_eigenvectors=True)
        elif k_max < 0:
            # Successively find more eigenvalues until some not close to zero
            # Number of new null space vectors to find in each step
            k_step = abs(k_max)
            k_max = min(A.shape[0] - 2, k_step)
            while k_max < A.shape[0] - 1:
                evals, evecs = sla.eigsh(A, sigma=sigma, which='LM',
                                         k=k_max, return_eigenvectors=True)
                if np.max(np.abs(evals)) > atol:
                    # We found the first large one
                    break
                else:
                    # All small, need to increase k_max
                    k_max += k_step
            else:
                raise ValueError('A should have at most A.shape[0]-2 dimensional null space.'
                                 'Try using sparse=False.')
        else:
            raise ValueError('k_max must be nonzero.')

        # Only keep eigenvectors with small eigenvalues
        nnz = np.isclose(evals, 0, atol=atol)
        ns = evecs[:, nnz]
        # Orthonormalize null space
        if ns.shape[1] > 0:
            ns, _ = la.qr(ns, mode='economic')
        return ns.T
    else:
        if isinstance(A, scipy.sparse.spmatrix):
            A = A.A
        # Do dense SVD
        u, s, vh = la.svd(A, full_matrices = True)
        nnz = np.isclose(s, 0, atol=atol)
        # Make sure it works for arbitrary rectangular matrices
        if len(s) < len(vh):
            nnz = np.concatenate((nnz, [True for _ in range(len(vh) - len(s))]))
        ns = vh[nnz].conj()
        nsc = vh[np.invert(nnz)].conj()
        if return_complement:
            return ns, nsc
        else:
            return ns


def split_list(vals, tol=1e-6):
    # Returns start and end indices of blocks of
    # consecutive close values in a list
    boundaries = np.where(np.abs(np.diff(vals)) > tol/len(vals))[0] + 1
    return np.array([np.insert(boundaries, 0, 0),
                     np.append(boundaries, len(vals))]).T


def simult_diag(mats, tol=1e-6, checks=0):
    """
    Simultaneously diagonalize commuting normal square matrices.
    Recursive algorithm, first diagonalize first matrix, express the
    rest of the matrices in its eigenbasis, and repeat the procedure
    for the rest of the matrices in each of the degenerate eigensubspaces.
    Recursion depth is limited by the number of nontrivial nested
    eigensubspaces, which is at most the number of matrices or
    the size of the matrices.

    Parameters:
    -----------------
    mats : ndarray
        List of commuting matrices, 3 index array.
    tol : float
        Tolerance when deciding degeneracy of eigenvalues.
    checks : int
        Amount of checks to do.
        0: no checks
        1: check the final result
        2: check all the intermediate results

    Returns:
    ---------------
    Ps : list of ndarrays
        List of rectangular matrices of common eigenvectors spanning each of
        the common eigensubspaces. Stacking them gives unitary matrix 'U = np.hstack(Ps)'
        such that 'U.T.conj() @ mats[i] @ U' is diagonal for every 'i'.

    Notes:
    It is recommended to order matrices such that those with eigenvalues that
    are far spaced or exactly degenerate (such as symmetries) are first, and
    those which may have accidental near degeneracies (such as Hamiltonians)
    are last.
    Tolerance is divided by the size of the block when deciding which eigenvalues
    are degenerate. The default value seems to work well, but not extensively
    tested, numerical instabilities are possible.
    """
    def grouped_diag(H, tol=1e-6):
        # Diagonalize normal matrix H and group the eigenvalues and eigenvectors
        # such that approximately equal eigenvalues are grouped together.
        # If H is Hermitian, 'eigh' returns ordered eigenvalues.
        # Returns the grouped eigenvalues, eigenvectors
        Hdag = H.T.conj()
        if allclose(H, Hdag):
            evals, U = la.eigh(H)
        else:
            if not np.allclose(commutator(H, Hdag), 0):
                raise ValueError('Only normal matrix can be diagonalized.')
            evals, U = la.eig(H)
            # Treat complex eigenvalues as 2D vectors
            evvec = np.array([evals.real, evals.imag]).T
            # Find connected clusters of close values
            con = cdist(evvec, evvec) < tol/len(H)
            _, groups = connected_components(con)
            # reorder evals and evecs such that groups are together
            order = np.argsort(groups)
            evals = evals[order]
            U = U[:, order]
            # Round U to unitary using QR, this only mixes vectors
            # from degenerate eigensubspaces
            U, _ = la.qr(U)
        if checks == 2:
            # Check the result
            assert allclose((U.T.conj() @ H @ U), np.diag(evals))
        return evals, U

    # If 1x1 matrix, we are done
    if mats[0].shape == (1, 1):
        return [np.eye(1)]

    # Diagonalize mats[0], if there are no more matrices we are done
    evals, U = grouped_diag(mats[0], tol=tol)
    ind = split_list(evals, tol=tol)
    if len(mats) == 1:
        return [U[:, b:e] for b, e in ind]

    # Check that all matrices commute with mats[0]
    if not np.allclose([commutator(mats[0], mat) for mat in mats[1:]], 0):
        raise ValueError('Only commuting matrices can be simultaneously diagonalized.')

    # Transform the rest of the matrices mats[1:] to the eigenbasis
    # of mats[0] where they are all block-diagonal.
    # Apply the same algorithm to the rest of the matrices in
    # each block recursively.
    matr = mtm(U.T.conj(), mats[1:], U)
    if checks == 2:
        # Check that off-diagonal blocks are small after the transformation
        assert np.all([np.allclose(matr[:, b1:e1, b2:e2], 0)
                for (b1, e1), (b2, e2) in it.combinations(ind, 2)])
    Ps = []
    for b, e in ind:
        P0 = U[:, b:e]
        Pnew = simult_diag(matr[:, b:e, b:e], tol=tol, checks=(2 if checks==2 else 0))
        Ps += [np.dot(P0,P) for P in Pnew]
    if checks > 0:
        # Check the result is diagonal
        U = np.hstack(Ps)
        matsd = mtm(U.T.conj(), mats, U)
        assert np.all([allclose(m, np.diag(np.diagonal(m))) for m in matsd])
    return Ps


def solve_mat_eqn(HL, HR=None, hermitian=False, traceless=False,
                  conjugate=False, sparse=None, k_max=-10):
    """Solve for X the simultaneous matrix equatioins X HL[i] = HR[i] X for every i.
    It is mapped to a system of linear equations, the null space of which gives a basis for
    all sulutions.

    Parameters:
    -----------------
    HL : ndarray or list of arrays or sparse matrices
        Coefficient matrices of identical square shape, list of arrays of
        shape (n, n) or one array of shape (m, n, n).
    HR : ndarray or list of ndarrays or None
        Same as HL. If HR = None, HR = HL is used.
    hermitian, traceless : booleans
        If X has to be hermitian, use hermitian = True.
        If X has to be traceless, use traceless = True.
    conjugate : boolean or list of booleans
        If True, solve X HL[i] = HR[i] X^* instead.
        If a list with the same length as HL and HR is provided, conjugation
        is applied to the equations with index corresponding to the True entries.
    sparse : bool  or None (default)
        Whether to use sparse linear algebra to find the null space.
        The default value is inferred from the type of the input array.
    k_max : int
        If k_max<0, all solutions are found by increasing the number
        of soulutions sought in each step by abs(k_max).
        If k_max>0, k_max is the maximum number of solutions requested.
        If the solution space dimension is higher than k_max, basis of a random
        subspace is returned. Can be used to increase performance if the
        maximum number of solutions is known.
        Ignored unless sparse is True.

    Returns:
    ---------------
    ndarray of shape (l, n, n), or list of csr_matrix
        List of linearly independent square matrices
        that span all solutions of the eaquations.
    """
    if sparse is None:
        sparse = isinstance(HL[0], scipy.sparse.spmatrix)
    if HR is None:
        HR = HL
    if len(HL) != len(HR) or not all(l.shape == r.shape for l, r in zip(HL, HR)):
        raise ValueError('HL and HR must have the same shape.')
    if isinstance(conjugate, bool):
        conjugate = [conjugate] * len(HL)
    if len(conjugate) != len(HL):
        raise ValueError('Conugate must have the same length as HL.')
    if not all(l.shape[0] == l.shape[1] == r.shape[0] == r.shape[1] for l, r in zip(HL, HR)):
        raise ValueError('HL and HR must be a list of square matrices.')

    dim = HL[0].shape[-1]
    # number of basis matrices
    N = dim**2 - (1 if traceless else 0)
    if N == 0:
        return np.empty((0, dim, dim))

    # Prepare for differences in sparse and dense algebra
    if sparse:
        basis = lambda: matrix_basis(dim, traceless=traceless, sparse=True)
        vstack = scipy.sparse.vstack
        bmat = scipy.sparse.bmat
        # Cast it to coo format and reshape
        flatten = lambda x: scipy.sparse.coo_matrix(x).reshape((x.shape[0]*x.shape[1], 1))
    else:
        basis = lambda: matrix_basis(dim, traceless=traceless, sparse=False)
        vstack = np.vstack
        bmat = np.block
        flatten = lambda x: x.reshape((-1, 1))

    # Calculate coefficients from commutators of the basis matrices
    null_mat = []
    for hL, hR, conj in zip(HL, HR, conjugate):
        if conj:
            print(next(basis))
            row = [flatten(mat @ hL - hR @ mat.conj()) for mat in basis()]
        else:
            row = [flatten(mat @ hL - hR @ mat) for mat in basis()]
        null_mat.append(row)

    null_mat = bmat(null_mat)

    if hermitian:
        # Simultaneaous equations for real and immaginary parts
        # Lapack guarantees that the SVD of a real matrix is real
        null_mat = vstack((null_mat.real, null_mat.imag))

    # Find the null space of null_mat
    ns = nullspace(null_mat, sparse=sparse, k_max=k_max)
    # Make all solutions
    # 'ij,jkl->ikl'
    basis = lambda: matrix_basis(dim, traceless=traceless, sparse=False)
    return np.array([sum((v * mat for v, mat in zip(vec, basis()))) for vec in ns])


def rref(m, rtol = 1e-3, return_S=False):
    """ Bring a matrix to reduced row echelon form.

    Parameters
    ----------
    m: 2D numpy.ndarray
        The matrix on which to perform row reduction.
    rtol: float
        The tolerance relative to the largest matrix element
        for what is considered a zero entry.
    return_S: bool
        Whether to return the matrix of linear operations
        that brings the input matrix to reduced row echelon form.

    Returns
    ---------
    red: 2D numpy.ndarray
        The input matrix in reduced row echelon form.
    S: 2D numpy.ndarray
        The matrix of operations that brings the input matrix
        to reduced row echelon form. Only returned if return_S = True.

    Note: If r is the reduced row echelon form of the input matrix m, then
        m = S.dot(r).
    """

    red = np.array(m)
    rows, columns = red.shape

    # Define row operations
    def pivot(i, j):
        P = np.eye(rows, dtype=complex)
        P[i, i] = P[j, j] = 0
        P[i, j] = P[j, i] = 1
        return P

    def rowsub(i, coeffs):
        P = np.eye(rows, dtype=complex)
        P[:, i] -= coeffs
        P[i, i] = 1
        return P

    def rowmul(i, coeff):
        P = np.eye(rows, dtype=complex)
        P[i, i] = coeff
        return P

    # Stop if m is identically zero
    if np.allclose(red, 0):
        return red

    S = np.eye(rows, dtype=complex)
    tol = np.max(np.abs(m)) * rtol

    lead = 0
    for r in range(rows):
        # find leading value below row r in lead column
        while lead < columns:
            i = np.argmax(np.abs(red[r:, lead])) + r
            lv = red[i, lead]
            if np.abs(lv) < tol:
                lead += 1
                continue
            else:
                break
        else:
            break

        T = pivot(i, r)
        S = T.dot(S)
        red = T.dot(red)
        T = rowmul(r, 1/lv)
        S = T.dot(S)
        red = T.dot(red)
        T = rowsub(r, red[:, lead])
        S = T.dot(S)
        red = T.dot(red)
        lead += 1
    if return_S:
        return red, S
    else:
        return red


def sparse_basis(bas, num_digits=3, reals=False):
    """Reduce the number of nonzero entries in a (basis) matrix by
    rounding the matrix and bringing it to reduced row echelon form.

    Parameters
    ----------
    bas: numpy.ndarray
        The (basis) matrix to prettify and round.
    num_digits: positive integer
        Number of significant digits to which to round.
    reals: boolean
        If True, the real and imaginary parts of the matrix
        are treated separately.

    Returns:
    ----------
    A rounded, sparsified matrix with the same row span as the input matrix.

    This is an attempt at matrix sparsification, i.e. we attempt to construct
    the sparsest matrix that has the same row rank. Note that the reduction to
    row echelon form is numerically unstable, so this function should be used
    with caution. """

    if len(bas) < 1:
        return bas
    if reals:
        re = np.real(bas)
        im = np.imag(bas)
        bas = np.hstack((re, im))
    bas_mat = rref(bas, rtol=10**(-num_digits))
    if reals:
        re, im = np.split(bas_mat, 2, axis=1)
        bas_mat = re + 1j * im
    # Round to num_digits and return only nonvanishing rows
    bas_mat = np.round(bas_mat, num_digits)
    bas_mat = np.vstack([row for row in bas_mat if not np.allclose(row, 0, atol=10**(-num_digits))])
    if len(bas_mat) < len(bas):
        warnings.warn('Removed linearly dependent terms from the family during sparsification. '+ \
                      'Resulting family will contain fewer members.')
    return np.round(bas_mat, num_digits)


def _inv_int(A):
    """Invert an integer square matrix A. """
    _A = ta.array(A, int)
    if A == np.empty((0, 0)):
        return A
    if _A != A or abs(la.det(A)) != 1:
        raise ValueError('Input needs to be an invertible integer matrix')
    return ta.array(np.round(la.inv(_A)), int)



import pytest
import warnings
import sympy
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import LinearOperator
import numpy as np

from pytest import raises
from ..model import _to_bloch_coeff, _commutative_momenta, Model, \
                    e, I, BlochModel, BlochCoeff, _symbol_normalizer
from ..linalg import allclose


k_x, k_y = _commutative_momenta[:2]
momenta = _commutative_momenta[:2]
c0, c1 = sympy.symbols('c0 c1', real=True)

def test_dense_algebra():
    # scalar models
    a = Model({1: 1, 'x': 3})
    assert a.shape == ()
    assert a.format is np.complex128
    assert a * 0 == 0 * a == a - a == 0
    assert 0 * a == a.zeros_like()
    assert 0 * a == {}
    assert a + 0 == a
    assert 0 + a == a
    assert a - a == {}
    assert a + a == 2 * a
    assert a + 2 == Model({1: 3, 'x': 3})
    assert a / 2 == a * 0.5
    a += a
    assert a == Model({1: 2, 'x': 6})
    a /= 2
    assert a == Model({1: 1, 'x': 3})
    b = Model({1: 2, 'x**2': 3})
    assert a + b == Model({1: 3, 'x': 3, 'x**2': 3})
    assert a * b == Model({1: 2, 'x': 6, 'x**2': 3, 'x**3': 9})
    assert (a - a) * b == {}
    assert (a - a) * (b - b) == {}
    # vector models
    vec = np.ones((3,))
    c = a * vec
    assert c.shape == (3,)
    assert c.format == np.ndarray
    assert c + 0 == c
    assert 0 + c == c
    assert c + 2 * vec == Model({1: 3, 'x': 3}) * vec
    assert 2 * vec + c == Model({1: 3, 'x': 3}) * vec
    # matrix models
    mat = np.ones((3, 3))
    d = b * mat
    assert d.shape == (3, 3)
    assert d.format == np.ndarray
    assert d * 0 == 0 * d == d - d == 0
    assert d + 0 == d
    assert 0 + d == d
    assert d @ c == 3 * Model({1: 2*vec, 'x': 6*vec, 'x**2': 3*vec, 'x**3': 9*vec})
    assert c.T() @ d == 3 * a * b
    assert c.T() @ d @ c == 9 * a * a * b
    assert d @ d == 3 * b * d
    assert (d @ d).format == np.ndarray
    assert d + 2 * mat == Model({1: 4, 'x**2': 3}) * mat
    assert 2 * mat + d == Model({1: 4, 'x**2': 3}) * mat
    # numpy elemntwise multiplication
    assert d * c == a * b * mat
    assert d.trace() == 3 * b
    assert d.reshape(-1) == b * np.ones((9,))


def test_dense_algebra_bloch():
    # scalar models
    a = BlochModel({1: 1, 'e**(I*k_x)': 3})
    assert a.shape == ()
    assert a.format is np.complex128
    assert a * 0 == 0 * a
    assert 0 * a == a.zeros_like()
    assert 0 * a == {}
    assert a - a == {}
    assert a + a == 2 * a
    assert a + 2 == BlochModel({1: 3, 'e**(I*k_x)': 3})
    assert a / 2 == a * 0.5
    a += a
    assert a == BlochModel({1: 2, 'e**(I*k_x)': 6})
    a /= 2
    assert a == BlochModel({1: 1, 'e**(I*k_x)': 3})
    b = BlochModel({1: 2, 'e**(I*2*k_x)': 3})
    assert a + b == BlochModel({1: 3, 'e**(I*2*k_x)': 3, 'e**(I*k_x)': 3})
    assert a * b == BlochModel({1: 2, 'e**(I*k_x)': 6, 'e**(I*2*k_x)': 3, 'e**(I*3*k_x)': 9})
    assert (a - a) * b == {}
    assert (a - a) * (b - b) == {}
    # vector models
    vec = np.ones((3,))
    c = a * vec
    assert c.shape == (3,)
    assert c.format == np.ndarray
    assert c + 2 * vec == BlochModel({1: 3, 'e**(I*k_x)': 3}) * vec
    # matrix models
    mat = np.ones((3, 3))
    d = b * mat
    assert d.shape == (3, 3)
    assert d.format == np.ndarray
    assert d @ c == 3 * BlochModel({1: 2*vec, 'e**(I*k_x)': 6*vec,
                                    'e**(I*2*k_x)': 3*vec, 'e**(I*3*k_x)': 9*vec})
    assert c.T() @ d == 3 * a * b
    assert c.T() @ d @ c == 9 * a * a * b
    assert d @ d == 3 * b * d
    assert (d @ d).format == np.ndarray
    assert d + 2 * mat == BlochModel({1: 4, 'e**(I*2*k_x)': 3}) * mat
    # numpy elemntwise multiplication
    assert d * c == a * b * mat
    assert d.trace() == 3 * b
    assert d.reshape(-1) == b * np.ones((9,))


def test_sparse_algebra():
    # Test sparse matrices
    a = Model({1: 1, 'x': 3})
    b = Model({1: 2, 'x**2': 3})
    # sparse vector model
    vec = csr_matrix(np.ones((3, 1)))
    c = a * vec
    assert c.shape == (3, 1)
    assert c.format is csr_matrix
    assert c + 2 * vec == Model({1: 3, 'x': 3}) * vec
    assert 2 * vec + c == Model({1: 3, 'x': 3}) * vec
    assert c + 0 == c
    assert 0 + c == c
    mat = csr_matrix(np.ones((3, 3)))
    d = b * mat
    assert d.shape == (3, 3)
    assert d.format is csr_matrix
    assert d + 0 == d
    assert 0 + d == d
    assert d @ c == 3 * Model({1: 2*vec, 'x': 6*vec, 'x**2': 3*vec, 'x**3': 9*vec})
    assert d @ d == 3 * b * d
    assert (d @ d) @ c == 9 * b * b * c
    assert (d @ d).format is csr_matrix
    assert d + 2 * mat == Model({1: 4, 'x**2': 3}) * mat
    assert 2 * mat + d == Model({1: 4, 'x**2': 3}) * mat
    assert c.T() @ d == 3 * b * c.T()
    assert c.T() @ d @ c == 9 * a * a * b * np.eye(1)
    assert d.trace() == 3 * b
    assert d.reshape((1, 9)) @ np.ones((9,)) == 9 * b
    e = d @ np.ones((3,))
    assert e == 3 * b * np.ones((3,))
    assert e.format is np.ndarray

    # Test LinearOperator
    d = b * LinearOperator((3, 3), matvec=lambda v: mat @ v)
    c = a * np.ones((3, 1))
    assert d.shape == (3, 3)
    assert d.format is LinearOperator
    assert d @ d == 3 * b * d
    assert (d @ d) @ c == 9 * b * b * c
    assert (d @ d).format is LinearOperator
    assert (d + 2 * LinearOperator((3, 3), matvec=lambda v: mat @ v)
            == Model({1: 4, 'x**2': 3}) * mat)
    assert d @ c == 3 * Model({1: 2*vec, 'x': 6*vec, 'x**2': 3*vec, 'x**3': 9*vec})
    assert (d @ c).format is np.ndarray
    assert c.T() @ (d @ c) == 9 * a * a * b


def test_sparse_algebra_bloch():
    # Test sparse matrices
    a = BlochModel({1: 1, 'e**(I*k_x)': 3})
    b = BlochModel({1: 2, 'e**(I*2*k_x)': 3})
    # sparse vector model
    vec = csr_matrix(np.ones((3, 1)))
    c = a * vec
    assert c.shape == (3, 1)
    assert c.format is csr_matrix
    assert c + 2 * vec == BlochModel({1: 3, 'e**(I*k_x)': 3}) * vec
    mat = csr_matrix(np.ones((3, 3)))
    d = b * mat
    assert d.shape == (3, 3)
    assert d.format is csr_matrix
    assert d @ c == 3 * BlochModel({1: 2*vec, 'e**(I*k_x)': 6*vec,
                                    'e**(I*2*k_x)': 3*vec, 'e**(I*3*k_x)': 9*vec})
    assert d @ d == 3 * b * d
    assert (d @ d) @ c == 9 * b * b * c
    assert (d @ d).format is csr_matrix
    assert d + 2 * mat == BlochModel({1: 4, 'e**(I*2*k_x)': 3}) * mat
    assert c.T() @ d == 3 * b * c.T()
    assert c.T() @ d @ c == 9 * a * a * b * np.eye(1)
    assert d.trace() == 3 * b
    assert d.reshape((1, 9)) @ np.ones((9,)) == 9 * b
    e = d @ np.ones((3,))
    assert e == 3 * b * np.ones((3,))
    assert e.format is np.ndarray

    # Test LinearOperator
    d = b * LinearOperator((3, 3), matvec=lambda v: mat @ v)
    c = a * np.ones((3, 1))
    assert d.shape == (3, 3)
    assert d.format is LinearOperator
    assert d @ d == 3 * b * d
    assert (d @ d) @ c == 9 * b * b * c
    assert (d @ d).format is LinearOperator
    assert (d + 2 * LinearOperator((3, 3), matvec=lambda v: mat @ v)
            == BlochModel({1: 4, 'e**(I*2*k_x)': 3}) * mat)
    assert d @ c == 3 * BlochModel({1: 2*vec, 'e**(I*k_x)': 6*vec,
                                    'e**(I*2*k_x)': 3*vec, 'e**(I*3*k_x)': 9*vec})
    assert (d @ c).format is np.ndarray
    assert c.T() @ (d @ c) == 9 * a * a * b


def test_to_bloch_coeff():

    key = sympy.sqrt(5)*e**(2*c0)*e**(I*(k_x/2 + np.sqrt(3)*k_y + c1))
    bc = _to_bloch_coeff(key, momenta)
    assert allclose(bc[0], np.array([1./2, np.sqrt(3)]))
    assert sympy.simplify(bc[1]) == sympy.simplify(sympy.sqrt(5)*e**(2*c0 + I*c1))

    key = sympy.sqrt(3)*c0
    bc = _to_bloch_coeff(key, momenta)
    assert allclose(bc[0], [0, 0])
    assert bc[1] == sympy.sqrt(3)*c0

    key = e**(2*k_x)
    with raises(ValueError):
        bc = _to_bloch_coeff(key, momenta)

    key = (c1 + 3)*e**I*(k_x + k_y)
    with raises(ValueError):
        bc = _to_bloch_coeff(key, momenta)

    key = c0*k_y*e**(I*k_x)
    with raises(ValueError):
        bc = _to_bloch_coeff(key, momenta)

    key = sympy.sqrt(5)*e**(2*c0)*e**(I*(k_x/2 + np.sqrt(3)*k_y + c1*k_y))
    with raises(ValueError):
        bc = _to_bloch_coeff(key, momenta)


def test_BlochCoeff():

    BC1 = BlochCoeff(np.array([2, 1]), c0)
    BC2 = BlochCoeff(np.array([3, -7]), c1)
    prod = BC1 * BC2
    assert allclose(prod[0], BC1[0] + BC2[0])
    assert prod[1] == BC1[1]*BC2[1]

    prod = BC1 * sympy.sqrt(3) * c1
    assert allclose(prod[0], BC1[0])
    assert prod[1] == sympy.sqrt(3)*BC1[1] * c1
    with raises(NotImplementedError):
        prod = BC1 * 3.1


def test_Model():

    m = Model("sqrt(3) * e^(I*k_x + 2*I*k_y) + e^(-I*k_y)/2")
    keys= [_symbol_normalizer(e**(I*k_x + 2*I*k_y)),
           _symbol_normalizer(e**(-I*k_y))]
    assert all([key in keys for key in m.keys()])
    assert allclose(m[keys[0]], np.sqrt(3))
    assert allclose(m[keys[1]], 0.5)

    np.random.seed(seed=0)
    mat = np.random.rand(2,2)
    m = Model({np.sqrt(5)*e**(I*k_x) : mat}, normalize=True)
    assert allclose(m[list(m.keys())[0]], mat*np.sqrt(5))

    m1 = Model({k_x : 2.3})
    m2 = Model({k_y : 2.4, k_x : 1})
    assert not m1 == m2
    m2 += m1
    assert m2 == Model({k_x : 3.3, k_y : 2.4})
    m2 = Model(m1, momenta=['k_x'])
    with raises(ValueError):
        m2 + m1
    m2 = m1
    assert m1 == m2

    m2 = Model({k_y : 2.4, k_x : 1})
    prod = m1 * m2
    keys = [k_x**2, k_x*k_y]
    assert all([key in keys for key in prod.keys()])
    assert prod[keys[0]] == m1[k_x] * m2[k_x]
    assert prod[keys[1]] == m1[k_x] * m2[k_y]
    assert (m1 * np.array([2]))[k_x] == 2*m1[k_x]

    m2.momenta = (k_x,)
    with raises(ValueError):
        m1*m2

    m1 = Model({k_x : 4j, e**(I*k_y): 5})
    keys = [k_x, e**(-I*k_y)]
    m2 = m1.conj()
    assert all([key in keys for key in m2.keys()])
    assert m2[k_x] == m1[k_x].conj()
    assert m2[e**(-I*k_y)] == m1[e**(I*k_y)].conj()


def test_BlochModel():

    m = Model({e**(I*k_y): 3*np.eye(2), k_x : np.eye(2)}, normalize=True)
    with raises(ValueError):
        bm = BlochModel(m)
    m = Model({c1*e**(I*k_y): 3*np.eye(2), np.sqrt(3)*e**(I*k_x) : np.eye(2)},
              momenta=['k_x', 'k_y'], normalize=True)
    bm = BlochModel(m)
    keys = [BlochCoeff(np.array([0, 1]), c1), BlochCoeff(np.array([1, 0]), sympy.numbers.One())]
    assert all([key in keys for key in bm.keys()])
    assert allclose(bm[keys[0]], m[c1*e**(I*k_y)])
    assert allclose(bm[keys[1]], m[e**(I*k_x)])

    bm2 = (bm + bm).tomodel(nsimplify=True)
    keys2 = [c1*e**(I*k_y), e**(I*k_x)]
    assert all([key in keys2 for key in bm2.keys()])
    assert allclose(bm2[keys2[0]], 2*bm[keys[0]])
    assert allclose(bm2[keys2[1]], 2*bm[keys[1]])


def test_Model_subs():
    np.random.seed(seed=0)
    T = np.random.randint(10, size=(2,2))
    c0, c1 = sympy.symbols('c0 c1', real=True)
    Ham = Model({c0 * e**(-I*(k_x/2 + k_y ) + sympy.sqrt(2)) : T,
                 c1 * e**(I*(4*k_x + 3*k_y)) : 2*T}, momenta=['k_x', 'k_y'])
    u_1, u_2 = sympy.symbols('u_1 u_2', real=True)
    nHam = Ham.subs({k_x: u_1, k_y: 2*k_y + 1})
    assert nHam.momenta == (u_1, k_y)
    right_Ham = Model({c0*e**(-2*I*k_y - I*u_1/2): T * np.exp(np.sqrt(2)) * np.exp(-1j),
                       c1*e**(6*I*k_y + 4*I*u_1): 2 * T *np.exp(3j)},
                      momenta=[u_1, k_y])
    assert right_Ham.allclose(nHam), list((right_Ham - nHam).keys())

    nHam = Ham.subs(k_y, 1.5)
    assert nHam.momenta == (k_x,)
    right_Ham = Model({c0*e**(-I*k_x/2): T * np.exp(np.sqrt(2)) * np.exp(-1.5j),
                       c1*e**(4*I*k_x): 2 * T * np.exp(3*1.5j)},
                      momenta=[u_1, k_y])
    assert right_Ham.allclose(nHam), list((right_Ham - nHam).keys())

    nHam = Ham.subs(k_y, sympy.sqrt(3) + u_1)
    assert nHam.momenta == (k_x,)

    T = np.random.randint(10, size=(2,2))
    Ham = BlochModel({c0 * e**(-I*(k_x/2 + k_y )) : T,
                          c1 * e**(I*(4*k_x + 3*k_y)) : 2*T}, momenta=['k_x', 'k_y'])
    nHam = Ham.subs([(c0, 3), (c1, 2*u_1)])
    right_Ham = BlochModel({BlochCoeff(np.array([-0.5, -1]), sympy.numbers.One()): T * 3,
                       BlochCoeff(np.array([4, 3]), u_1): 2 * T * 2},
                      momenta=[k_x, k_y])

    assert right_Ham.allclose(nHam), list((right_Ham - nHam).keys())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
import sys


if sys.version_info < (3, 5):
    print('Qsymm requires Python 3.5 or above.')
    sys.exit(1)


install_requires = [
    'numpy>=1.13',  # because we use __array_ufunc__
    'scipy>=1.1.0', # because we use reshape for sparse matrices
    'sympy>=1.1',
    'tinyarray',
]

extras_require = {
    'kwant': ['kwant'],
}


# Loads version.py module without importing the whole package.
def get_version_and_cmdclass(package_path):
    import os
    from importlib.util import module_from_spec, spec_from_file_location
    spec = spec_from_file_location('version',
                                   os.path.join(package_path, '_version.py'))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


version, cmdclass = get_version_and_cmdclass('qsymm')


setup(
    name='qsymm',
    description='Symmetry finder and symmetric Hamiltonian generator',
    url='https://gitlab.kwant-project.org/qt/qsymm',
    author='Qsymm authors',
    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.5',
    ],
    packages=find_packages('.'),
    install_requires=install_requires,
    extras_require=extras_require,
    version=version,
    cmdclass=cmdclass,
)



import pytest
import numpy as np
import scipy.linalg as la
from copy import deepcopy
import itertools as it

from ..groups import generate_group, generate_subgroups
from ..symmetry_finder import discrete_symmetries
from ..model import Model
from ..groups import PointGroupElement, spin_matrices, spin_rotation, L_matrices
from ..hamiltonian_generator import continuum_hamiltonian
from ..linalg import prop_to_id
from ..kwant_continuum import sympify

sigma = spin_matrices(1/2)

def test_mutual_continuum():
    # Tests that check continuum hamiltonian generator against symmetry finder

    # 0D
    dim = 0
    for N in np.arange(2, 5):
        # Symplistic symmetry action
        TR = PointGroupElement(np.eye(dim), True, False, np.eye(N))
        PH = PointGroupElement(np.eye(dim), True, True, np.eye(N))
        gens = {TR, PH}
        group = generate_group(gens)
        groupnoU = deepcopy(group)
        for g in groupnoU:
            g.U = None
        subgroups = generate_subgroups(group)
        for sg, gen in subgroups.items():
            families = continuum_hamiltonian(list(gen), dim, 3)
            if not len(families) == 0:
                H = Model({sympify('a_' + str(i)) * k: v
                    for i, fam in enumerate(families) for k, v in fam.items()},
                    momenta=('k_x', 'k_y', 'k_z')[:dim])
                sg2, Ps = discrete_symmetries(H, groupnoU)
                # new symmetry group may bigger because of additional constraints
                assert sg2 >= sg, (sg2, sg)
                for g1, g2 in it.product(sg, sg2):
                    if g1 == g2:
                        prop, coeff = prop_to_id((g1.inv() * g2).U)
                        assert prop and np.isclose(abs(coeff), 1)

    # More realistic symmetry action
    TR = PointGroupElement(np.eye(dim), True, False, np.kron(sigma[1], np.eye(2)))
    PH = PointGroupElement(np.eye(dim), True, True, np.kron(np.eye(2), sigma[0]))
    gens = {TR, PH}
    group = generate_group(gens)
    groupnoU = deepcopy(group)
    for g in groupnoU:
        g.U = None
    subgroups = generate_subgroups(group)
    for sg, gen in subgroups.items():
        families = continuum_hamiltonian(list(gen), dim, 3)
        if not len(families) == 0:
            H = Model({sympify('a_' + str(i)) * k: v
                for i, fam in enumerate(families) for k, v in fam.items()},
                momenta=('k_x', 'k_y', 'k_z')[:dim])
            sg2, Ps = discrete_symmetries(H, groupnoU)
            # new symmetry group may bigger because of additional constraints
            assert sg2 == sg, (sg2, sg)
            for g1, g2 in it.product(sg, sg2):
                if g1 == g2:
                    prop, coeff = prop_to_id((g1.inv() * g2).U)
                    assert prop and np.isclose(abs(coeff), 1)

    # 1D
    dim = 1
    # More realistic symmetry action
    TR = PointGroupElement(np.eye(dim), True, False, np.kron(sigma[1], np.eye(2)))
    PH = PointGroupElement(np.eye(dim), True, True, np.kron(np.eye(2), sigma[2]))
    I = PointGroupElement(-np.eye(dim), False, False, np.eye(4))
    gens = {TR, PH, I}
    group = generate_group(gens)
    groupnoU = deepcopy(group)
    for g in groupnoU:
        g.U = None
    subgroups = generate_subgroups(group)
    for sg, gen in subgroups.items():
        families = continuum_hamiltonian(list(gen), dim, 1)
        if not len(families) == 0:
            H = Model({sympify('a_' + str(i)) * k: v
                for i, fam in enumerate(families) for k, v in fam.items()},
                momenta=('k_x', 'k_y', 'k_z')[:dim])
            sg2, Ps = discrete_symmetries(H, groupnoU)
            assert sg2 == sg, (sg2, sg)
            for g1, g2 in it.product(sg, sg2):
                if g1 == g2:
                    prop, coeff = prop_to_id((g1.inv() * g2).U)
                    assert prop and np.isclose(abs(coeff), 1)

    # 2D
    dim = 2
    L = L_matrices(2)
    # More realistic symmetry action
    TR = PointGroupElement(np.eye(dim), True, False, np.kron(sigma[1], np.eye(2)))
    PH = PointGroupElement(np.eye(dim), True, True, np.kron(np.eye(2), sigma[0]))
    M = PointGroupElement(-np.array([[-1, 0], [0, 1]]), False, False,
                         la.block_diag(spin_rotation(np.pi*np.array([1, 0, 0]), sigma),
                                        spin_rotation(np.pi*np.array([1, 0, 0]), sigma).conj()))
    n = np.pi * np.array([0, 0, 1/2])
    C4 = PointGroupElement(spin_rotation([n[2]], L, roundint=True), False, False,
                         la.block_diag(spin_rotation(n, sigma),
                                        spin_rotation(n, sigma).conj()))
    gens = {TR, PH, M, C4}
    group = generate_group(gens)
    groupnoU = deepcopy(group)
    for g in groupnoU:
        g.U = None
    families = continuum_hamiltonian(list(gens), dim, 1)
    if not len(families) == 0:
        H = Model({sympify('a_' + str(i)) * k: v
            for i, fam in enumerate(families) for k, v in fam.items()},
            momenta=('k_x', 'k_y', 'k_z')[:dim])
        sg2, Ps = discrete_symmetries(H, groupnoU)
        assert sg2 == group, (sg2, group)
        for g1, g2 in it.product(group, sg2):
            if g1 == g2:
                prop, coeff = prop_to_id((g1.inv() * g2).U)
                assert prop and np.isclose(abs(coeff), 1)



# -*- coding: utf-8 -*-

import numpy as np
import tinyarray as ta
import scipy.linalg as la
from itertools import product
from functools import lru_cache
from fractions import Fraction
from numbers import Number
from collections import OrderedDict
import sympy
from copy import deepcopy

from .linalg import prop_to_id, _inv_int, allclose
from .model import Model


# Cache the operations that are potentially slow and happen a lot in
# group theory applications. Essentially store the multiplication table.
@lru_cache(maxsize=10000)
def _mul(R1, R2):
    # Cached multiplication of spatial parts.
    if is_sympy_matrix(R1) and is_sympy_matrix(R2):
        # If spatial parts are sympy matrices, use cached multiplication.
        R = R1 * R2
    elif not (is_sympy_matrix(R1) or is_sympy_matrix(R2)):
        # If arrays, use dot
        R = ta.dot(R1, R2)
    elif ((is_sympy_matrix(R1) or is_sympy_matrix(R2)) and
        (isinstance(R1, ta.ndarray_int) or isinstance(R2, ta.ndarray_int))):
        # Multiplying sympy and integer tinyarray is ok, should result in sympy
        R = sympy.ImmutableMatrix(R1) * sympy.ImmutableMatrix(R2)
    else:
        raise ValueError("Mixing of sympy and floating point in the spatial part R is not allowed. "
                         "To avoid this error, make sure that all PointGroupElements are initialized "
                         "with either floating point arrays or sympy matrices as rotations. "
                         "Integer arrays are allowed in both cases.")
    R = _make_int(R)
    return R


@lru_cache(maxsize=1000)
def _inv(R):
    if isinstance(R, ta.ndarray_int):
        Rinv = ta.array(_inv_int(R), int)
    elif isinstance(R, ta.ndarray_float):
        Rinv = ta.array(la.inv(R))
    elif is_sympy_matrix(R):
        Rinv = R**(-1)
    else:
        raise ValueError('Invalid rotation matrix.')
    return Rinv


@lru_cache(maxsize=10000)
def _eq(R1, R2):
    if type(R1) != type(R2):
        R1 = ta.array(np.array(R1).astype(float), float)
        R2 = ta.array(np.array(R2).astype(float), float)
    if isinstance(R1, ta.ndarray_float):
        # Check equality with allclose if floating point
        R_eq = allclose(R1, R2)
    else:
        # If exact use exact equality
        R_eq = (R1 == R2)
    return R_eq


@lru_cache(maxsize=1000)
def _make_int(R):
    # If close to an integer array convert to integer tinyarray, else
    # return original array
    R_float = (np.array(R).astype(float) if is_sympy_matrix(R) else R)
    R_int = ta.array(np.round(R_float), int)
    if allclose(R_float, R_int):
        return R_int
    else:
        return R


def _is_hermitian(a):
    return allclose(a, a.conjugate().transpose())


def _is_antisymmetric(a):
    return allclose(a, -a.transpose())


def is_sympy_matrix(R):
    # Returns True if the input is a sympy.Matrix or sympy.ImmutableMatrix.
    return isinstance(R, (sympy.ImmutableMatrix, sympy.matrices.MatrixBase))


class PointGroupElement:
    """An element of a point group.

    Parameters
    ----------
    R : sympy.ImmutableMatrix or array
        Real space rotation action of the operator. Square matrix with size
        of the number of spatial dimensions.
    conjugate : boolean (default False)
        Whether the operation includes conplex conjugation (antiunitary operator)
    antisymmetry : boolean (default False)
        Whether the operator flips the sign of the Hamiltonian (antisymmetry)
    U : ndarray (optional)
        The unitary action on the Hilbert space.
        May be None, to be able to treat symmetry candidates
    _strict_eq : boolean (default False)
        Whether to test the equality of the unitary parts when comparing with
        other PointGroupElements. By default the unitary parts are ignored.
        If True, PointGroupElements are considered equal, if the unitary parts
        are proportional, an overall phase difference is still allowed.

    Notes
    -----
    As U is floating point and has a phase ambiguity at least, hence
    it is ignored when comparing objects by default.

    R is the real space rotation acion. Do not include minus sign for
    the k-space action of antiunitary operators, such as time reversal.
    This minus sign will be included automatically if 'conjugate=True'.

    For most uses R can be provided as a floating point array. It is
    necessary to use exact sympy matrix representation if the PGE has
    to act on Models with complicated momentum dependence (not polynomial),
    as the function parts of models are compared exactly. If the momentum
    dependence is periodic (sine, cosine and exponential), use BlochModel,
    this works with floating point rotations.
    """

    __slots__ = ('R', 'conjugate', 'antisymmetry', 'U', '_strict_eq')

    def __init__(self, R, conjugate=False, antisymmetry=False, U=None, _strict_eq=False):
        if isinstance(R, sympy.ImmutableMatrix):
            # If it is integer, recast to integer tinyarray
            R = _make_int(R)
        elif isinstance(R, ta.ndarray_int):
            pass
        elif isinstance(R, ta.ndarray_float):
            R = _make_int(R)
        elif isinstance(R, sympy.matrices.MatrixBase):
            R = sympy.ImmutableMatrix(R)
            R = _make_int(R)
        elif isinstance(R, np.ndarray):
            # If it is integer, recast to integer tinyarray
            R = ta.array(R)
            R = _make_int(R)
        else:
            raise ValueError('Real space rotation must be provided as a sympy matrix or an array.')
        self.R, self.conjugate, self.antisymmetry, self.U = R, conjugate, antisymmetry, U
        # Calculating sympy inverse is slow, remember it
        self._strict_eq = _strict_eq

    def __repr__(self):
        return ('\nPointGroupElement(\nR = {},\nconjugate = {},\nantisymmetry = {},\nU = {})'
                .format(repr(self.R).replace('\n', '\n    '),
                        self.conjugate,
                        self.antisymmetry,
                        repr(self.U).replace('\n', '\n    ') if self.U is not None else 'None'))

    def __str__(self):
        return pretty_print_pge(self, full=True)

    def _repr_latex_(self):
        return pretty_print_pge(self, full=False, latex=True)

    def _repr_pretty_(self, pp, cycle):
        pp.text(pretty_print_pge(self, full=False))

    def __eq__(self, other):
        R_eq = _eq(self.R, other.R)
        basic_eq = R_eq and ((self.conjugate, self.antisymmetry) == (other.conjugate, other.antisymmetry))
        # Equality is only checked for U if _strict_eq is True and basic_eq is True.
        if basic_eq and (self._strict_eq is True or other._strict_eq is True):
            if (self.U is None) and (other.U is None):
                U_eq = True
            elif (self.U is None) ^ (other.U is None):
                U_eq = False
            else:
                prop, coeff = prop_to_id((self.inv() * other).U)
                U_eq = (prop and np.isclose(abs(coeff), 1))
        else:
            U_eq = True
        return basic_eq and U_eq

    def __lt__(self, other):
        # Sort group elements:
        # First by conjugate and a, then R = identity, then the rest
        # lexicographically
        Rs = ta.array(np.array(self.R).astype(float), float)
        Ro = ta.array(np.array(other.R).astype(float), float)
        identity = ta.array(np.eye(Rs.shape[0], dtype=int))

        if not (self.conjugate, self.antisymmetry) == (other.conjugate, other.antisymmetry):
            return (self.conjugate, self.antisymmetry) < (other.conjugate, other.antisymmetry)
        elif (Rs == identity) ^ (Ro == identity):
            return Rs == identity
        else:
            return Rs < Ro

    def __hash__(self):
        # U is not hashed, if R is floating point it is also not hashed
        R, c, a = self.R, self.conjugate, self.antisymmetry
        if isinstance(R, ta.ndarray_float):
            return hash((c, a))
        else:
            return hash((R, c, a))

    def __mul__(self, g2):
        g1 = self
        R1, c1, a1, U1 = g1.R, g1.conjugate, g1.antisymmetry, g1.U
        R2, c2, a2, U2 = g2.R, g2.conjugate, g2.antisymmetry, g2.U

        if (U1 is None) or (U2 is None):
            U = None
        elif c1:
            U = U1.dot(U2.conj())
        else:
            U = U1.dot(U2)
        R = _mul(R1, R2)
        return PointGroupElement(R, c1^c2, a1^a2, U, _strict_eq=(self._strict_eq or g2._strict_eq))

    def __pow__(self, n):
        result = self.identity()
        g = (self if n >=0 else self.inv())
        for _ in range(abs(n)):
            result *= g
        return result

    def inv(self):
        """Invert PointGroupElement"""
        R, c, a, U = self.R, self.conjugate, self.antisymmetry, self.U
        if U is None:
            Uinv = None
        elif c:
            Uinv = la.inv(U).conj()
        else:
            Uinv = la.inv(U)
        # Check if inverse is stored, if not, calculate it
        Rinv = _inv(R)
        result = PointGroupElement(Rinv, c, a, Uinv, _strict_eq=self._strict_eq)
        return result

    def _strictereq(self, other):
        # Stricter equality, testing the unitary parts to be approx. equal
        if (self.U is None) or (other.U is None):
            return False
        return ((self.R, self.conjugate, self.antisymmetry) == (other.R, other.conjugate, other.antisymmetry) and
                allclose(self.U, other.U))

    def apply(self, model):
        """Return copy of model with applied symmetry operation.

        if unitary: (+/-) U H(inv(R) k) U^dagger
        if antiunitary: (+/-) U H(- inv(R) k).conj() U^dagger

        (+/-) stands for (symmetry / antisymmetry)

        If self.U is None, U is taken as the identity.
        """
        R, antiunitary, antisymmetry, U = self.R, self.conjugate, self.antisymmetry, self.U
        R = _inv(R)
        R = R * (-1 if antiunitary else 1)
        result = model.rotate_momenta(R)
        if antiunitary:
            result = result.conj()
        if antisymmetry:
            result = -result
        if U is not None:
            result = U @ result @ U.T.conj()

        return result

    def identity(self):
        """Return identity element with the same structure as self."""
        dim = self.R.shape[0]
        R = ta.identity(dim, int)
        if self.U is not None:
            U = np.eye(self.U.shape[0])
        else:
            U = None
        return PointGroupElement(R, False, False, U)

## Factory functions for point group elements

def identity(dim, shape=None):
    """Return identity operator with appropriate shape.

    Parameters
    ----------
    dim : int
        Dimension of real space.
    shape : int (optional)
        Size of the unitary part of the operator.
        If not provided, U is set to None.

    Returns
    -------
    id : PointGroupElement
    """
    R = ta.identity(dim, int)
    if shape is not None:
        U = np.eye(shape)
    else:
        U = None
    return PointGroupElement(R, False, False, U)


def time_reversal(realspace_dim, U=None, spin=None):
    """Return a time-reversal symmetry operator

    parameters
    ----------
    realspace_dim : int
        Realspace dimension
    U: ndarray (optional)
        The unitary action on the Hilbert space.
        May be None, to be able to treat symmetry candidates.
    spin : float or sequence of arrays (optional)
        Spin representation to use for the unitary action of the time reversal
        operator. If float is provided, it should be integer or half-integer
        specifying the spin representation in the standard basis, see `spin_matrices`.
        Otherwise a sequence of 3 arrays of identical square size must be provided
        representing 3 components of the angular momentum operator. The unitary action
        of time-reversal operator is `U = exp(-i π s_y)`. Only one of `U` and `spin`
        may be provided.

    Returns
    -------
    T : PointGroupElement
    """
    if U is not None and spin is not None:
        raise ValueError('Only one of `U` and `spin` may be provided.')
    if spin is not None:
        U = spin_rotation(np.pi * np.array([0, 1, 0]), spin)
    R = ta.identity(realspace_dim, int)
    return PointGroupElement(R, conjugate=True, antisymmetry=False, U=U)


def particle_hole(realspace_dim, U=None):
    """Return a particle-hole symmetry operator

    parameters
    ----------
    realspace_dim : int
        Realspace dimension
    U: ndarray (optional)
        The unitary action on the Hilbert space.
        May be None, to be able to treat symmetry candidates

    Returns
    -------
    P : PointGroupElement
    """
    R = ta.identity(realspace_dim, int)
    return PointGroupElement(R, conjugate=True, antisymmetry=True, U=U)


def chiral(realspace_dim, U=None):
    """Return a chiral symmetry operator

    parameters
    ----------
    realspace_dim : int
        Realspace dimension
    U: ndarray (optional)
        The unitary action on the Hilbert space.
        May be None, to be able to treat symmetry candidates

    Returns
    -------
    P : PointGroupElement
    """
    R = ta.identity(realspace_dim, int)
    return PointGroupElement(R, conjugate=False, antisymmetry=True, U=U)


def inversion(realspace_dim, U=None):
    """Return an inversion operator

    parameters
    ----------
    realspace_dim : int
        Realspace dimension
    U: ndarray (optional)
        The unitary action on the Hilbert space.
        May be None, to be able to treat symmetry candidates

    Returns
    -------
    P : PointGroupElement
    """
    R = -ta.identity(realspace_dim, int)
    return PointGroupElement(R, conjugate=False, antisymmetry=False, U=U)


def rotation(angle, axis=None, inversion=False, U=None, spin=None):
    """Return a rotation operator

    parameters
    ----------
    angle : float
        Rotation angle in units of 2 pi.
    axis : ndarray or None (default)
        Rotation axis, optional. If not provided, a 2D rotation is generated
        around the axis normal to the plane. If a 3D vector is provided,
        a 3D rotation is generated around this axis. Does not need to be
        normalized to 1.
    inversion : bool (default False)
        Whether to generate a rotoinversion. By default a proper rotation
        is returned. Only valid in 3D.
    U: ndarray (optional)
        The unitary action on the Hilbert space.
        May be None, to be able to treat symmetry candidates
    spin : float or sequence of arrays (optional)
        Spin representation to use for the unitary action of the
        operator. If float is provided, it should be integer or half-integer
        specifying the spin representation in the standard basis, see `spin_matrices`.
        Otherwise a sequence of 3 arrays of identical square size must be provided
        representing 3 components of the angular momentum operator. The unitary action
        of rotation operator is `U = exp(-i n⋅s)`. In 2D the z axis is assumed to be
        the rotation axis. Only one of `U` and `spin` may be provided.

    Returns
    -------
    P : PointGroupElement
    """
    if U is not None and spin is not None:
        raise ValueError('Only one of `U` and `spin` may be provided.')

    angle = 2 * np.pi * angle
    if axis is None:
        # 2D
        R = np.array([[np.cos(angle), -np.sin(angle)],
                      [np.sin(angle), np.cos(angle)]])
        if spin is not None:
            U = spin_rotation(angle * np.array([0, 0, 1]), spin)
    elif len(axis) == 3:
        # 3D
        n = angle * np.array(axis, float) / la.norm(axis)
        R = spin_rotation(n, L_matrices(d=3, l=1))
        R *= (-1 if inversion else 1)
        if spin is not None:
            U = spin_rotation(n, spin)
    else:
        raise ValueError('`axis` needs to be `None` or a 3D vector.')
    return PointGroupElement(R.real, conjugate=False, antisymmetry=False, U=U)


def mirror(axis, U=None, spin=None):
    """Return a mirror operator

    Parameters:
    -----------
    axis : ndarray
        Normal of the mirror. The dimensionality of the operator is the same
        as the length of `axis`.
    U: ndarray (optional)
        The unitary action on the Hilbert space.
        May be None, to be able to treat symmetry candidates
    spin : float or sequence of arrays (optional)
        Spin representation to use for the unitary action of the
        operator. If float is provided, it should be integer or half-integer
        specifying the spin representation in the standard basis, see `spin_matrices`.
        Otherwise a sequence of 3 arrays of identical square size must be provided
        representing 3 components of the angular momentum operator. The unitary action
        of mirror operator is `U = exp(-i π n⋅s)` where n is normalized to 1. In 2D the
        axis is treated as x and y coordinates. Only one of `U` and `spin` may be provided.

    Returns:
    --------
    P : PointGroupElement

    Notes:
    ------
        Warning: in 2D the real space action of a mirror and and a 2-fold rotation
        around an axis in the plane is identical, however the action on angular momentum
        is different. Here we consider the action of the mirror, which is the same as the
        action of a 2-fold rotation around the mirror axis.
    """
    if U is not None and spin is not None:
        raise ValueError('Only one of `U` and `spin` may be provided.')

    axis = np.array(axis, float)
    axis /= la.norm(axis)
    R = np.eye(axis.shape[0]) - 2 * np.outer(axis, axis)
    if spin is not None:
        if len(axis) == 2:
            axis = np.append(axis, 0)
        U = spin_rotation(np.pi * axis, spin)

    return PointGroupElement(R, conjugate=False, antisymmetry=False, U=U)

## Continuous symmetry generators (conserved quantities)

class ContinuousGroupGenerator:
    r"""A generator of a continuous group.

    Generates a family of symmetry operators that act on the Hamiltonian as:

    .. math:: H(k) → \exp{-iλU} H(\exp{iλR} k) \exp{iλU}

    with λ a real parameter.

    Parameters:
    -----------
    R: ndarray (optional)
        Real space rotation generator: Hermitian antisymmetric.
        If not provided, the zero matrix is used.
    U: ndarray (optional)
        Hilbert space unitary rotation generator: Hermitian.
        If not provided, the zero matrix is used.
    """

    __slots__ = ('R', 'U')

    def __init__(self, R=None, U=None):
        # Make sure that R and U have correct properties
        if R is not None and not _is_hermitian(R) and not _is_antisymmetric(R):
            raise ValueError('R must be Hermitian antisymmetric')
        if U is not None and not _is_hermitian(U):
            raise ValueError('U must be Hermitian')
        self.R, self.U = R, U

    def __repr__(self):
        return ('\nContinuousGroupGenerator(\nR = {},\nU = {})'
                .format(repr(self.R).replace('\n', '\n    ') if self.R is not None else 'None',
                        repr(self.U).replace('\n', '\n    ') if self.U is not None else 'None'))

    def __str__(self):
        return pretty_print_cgg(self, latex=False)

    def _repr_latex_(self):
        return pretty_print_cgg(self, latex=True)

    def _repr_pretty_(self, pp, cycle):
        pp.text(pretty_print_cgg(self))

    def apply(self, model):
        """Return copy of model `H(k)` with applied infinitesimal generator.
        1j * (H(k) U - U H(k)) + 1j * dH(k)/dk_i R_{ij} k_j
        """
        R, U = self.R, self.U
        momenta = model.momenta
        R_nonzero = not (R is None or allclose(R, 0))
        U_nonzero = not (U is None or allclose(U, 0))
        if R_nonzero:
            dim = R.shape[0]
            assert len(momenta) == dim
        result = model.zeros_like()
        if R_nonzero:
            def trf(key):
                return sum([sympy.diff(key, momenta[i]) * R[i, j] * momenta[j]
                          for i, j in product(range(dim), repeat=2)])
            result += 1j * model.transform_symbolic(trf)
        if U_nonzero:
            result += model @ (1j*U) + (-1j*U) @ model
        return result


## General group theory algorithms

def generate_group(gens):
    """Generate group from gens

    Parameters:
    -----------
    gens : iterable of PointGroupElement
        generator set of the group

    Returns:
    --------
    group : set of PointGroupElement
        group generated by gens, closed under multiplication
    """
    gens = set(gens.copy())
    # here we keep all the elements generated so far
    group = gens.copy()
    # these are the elements generated in the previous step
    oldgroup = gens.copy()
    while True:
        # generate new elements by multiplying old elements with generators
        newgroup = {a * b for a, b in product(oldgroup, gens)}
        # only keep those that are new
        newgroup -= group
        # if there are any new, add them to group and set them as old
        if len(newgroup) > 0:
            group |= newgroup
            oldgroup = newgroup.copy()
        # if there were no new elements, we are done
        else:
            break
    return set(group)


def set_multiply(G, H):
    # multiply sets of group elements
    return {g * h for g, h in product(G, H)}


def generate_subgroups(group):
    """Generate all subgroups of group, including the trivial group
    and itself.

    Parameters:
    -----------
    group : set of PointGroupElement
        A closed group as set of its elements.

    Returns:
    --------
    subgroups : dict of forzenset: set
        frozesets are the subgroups, sets are a generator set of the
        subgroup.
    """
    # Make all the cyclic subgroups generated by one element
    sg1 = {frozenset(generate_group({g})): {g} for g in group}
    # Iteratively generate all subgroups generated by more elements
    # here we keep all the subgroups generated so far
    subgroups = sg1.copy()
    # these are the subgroups generated in the previous step
    sgold = sg1.copy()

    while True:
        sgnew = dict()
        # extend subgroups from previous step by all cyclic groups
        for (sg, gen), (g1, gen1) in product(sgold.items(), sg1.items()):
            # if cyclic group is already contained in group,
            # no point extending by it
            if not g1 <= sg:
                newsg = frozenset(generate_group(gen | gen1))
                # don't do anything if it is already generated
                # with less or equal number of generators
                if newsg not in subgroups:
                    # If we managed to extend anything, we append
                    # it to the list of subgroups and new subgroups
                    newgen = gen | gen1
                    subgroups[newsg] = newgen
                    sgnew[newsg] = newgen
        if len(sgnew) > 0:
            sgold = sgnew.copy()
        # If no extension, or we are already at the full group, stop
        if (len(sgnew) == 0 or
            min([len(sg) for sg in sgnew.keys()]) == len(group)):
            break

    return subgroups


## Predefined point groups

def square(tr=True, ph=True, generators=False, spin=None):
    """
    Generate square point group in standard basis.

    Parameters:
    -----------
    tr, ph : bool (default True)
        Whether to include time-reversal and particle-hole
        symmetry.
    generators : bool (default false)
        Only return the group generators if True.
    spin : float or sequence of arrays (optional)
        Spin representation to use for the unitary action of the
        operator. If not provided, the PointGroupElements have the unitary
        action set to None. If float is provided, it should be integer or half-integer
        specifying the spin representation in the standard basis, see `spin_matrices`.
        Otherwise a sequence of 3 arrays of identical square size must be provided
        representing 3 components of the angular momentum operator. The unitary action
        of rotation operator is `U = exp(-i n⋅s)`. In 2D the z axis is assumed to be
        the rotation axis. If `ph` is True, `spin` may not be provided, as it is not
        possible to deduce the unitary representation of particle-hole symmetry from
        spin alone. In this case construct the particle-hole operator manually.

    Returns:
    --------
    set of PointGroupElement objects with integer rotations

    Notes:
    ------
        Warning: in 2D the real space action of a mirror and and a 2-fold rotation
        around an axis in the plane is identical, however the action on angular  momentum
        is different. Here we consider the action of the mirror, which is the same as the
        action of a 2-fold rotation around the mirror axis, assuming inversion acts trivially.
    """
    if ph and spin is not None:
        raise ValueError('If `ph` is True, `spin` may not be provided, as it is not '
                         'possible to deduce the unitary representation of particle-hole symmetry '
                         'from spin alone. In this case construct the particle-hole operator manually.')
    Mx = mirror([1, 0], spin=spin)
    C4 = rotation(1/4, spin=spin)
    gens = {Mx, C4}
    if tr:
        TR = time_reversal(2, spin=spin)
        gens.add(TR)
    if ph:
        PH = particle_hole(2)
        gens.add(PH)
    if generators:
        return gens
    else:
        return generate_group(gens)


def cubic(tr=True, ph=True, generators=False, spin=None):
    """
    Generate cubic point group in standard basis.

    Parameters:
    -----------
    tr, ph : bool (default True)
        Whether to include time-reversal and particle-hole
        symmetry.
    generators : bool (default false)
        Only return the group generators if True.
    spin : float or sequence of arrays (optional)
        Spin representation to use for the unitary action of the
        operator. If not provided, the PointGroupElements have the unitary
        action set to None. If float is provided, it should be integer or half-integer
        specifying the spin representation in the standard basis, see `spin_matrices`.
        Otherwise a sequence of 3 arrays of identical square size must be provided
        representing 3 components of the angular momentum operator. The unitary action
        of rotation operator is `U = exp(-i n⋅s)`. If `ph` is True, `spin` may not be
        provided, as it is not  possible to deduce the unitary representation of
        particle-hole symmetry from spin alone. In this case construct the
        particle-hole operator manually.

    Returns:
    --------
    set of PointGroupElement objects with integer rotations

    Notes:
    ------
        We assume inversion acts trivially in spin space.
    """
    if ph and spin is not None:
        raise ValueError('If `ph` is True, `spin` may not be provided, as it is not '
                         'possible to deduce the unitary representation of particle-hole symmetry '
                         'from spin alone. In this case construct the particle-hole operator manually.')
    I = inversion(3, U=(None if spin is None else spin_rotation(np.zeros(3), spin)))
    C4 = rotation(1/4, [1, 0, 0], spin=spin)
    C3 = rotation(1/3, [1, 1, 1], spin=spin)
    cubic_gens = {I, C4, C3}
    if tr:
        TR = time_reversal(3, spin=spin)
        cubic_gens.add(TR)
    if ph:
        PH = particle_hole(3)
        cubic_gens.add(PH)
    if generators:
        return cubic_gens
    else:
        return generate_group(cubic_gens)


def hexagonal(dim=2, tr=True, ph=True, generators=False, sympy_R=True, spin=None):
    """
    Generate hexagonal point group in standard basis in 2 or 3 dimensions.
    Mirror symmetries with the main coordinate axes as normals are included.
    In 3D the hexagonal axis is the z axis.

    Parameters:
    -----------
    dim : int (default 2)
        Real sapce dimensionality, 2 or 3.
    tr, ph : bool (default True)
        Whether to include time-reversal and particle-hole
        symmetry.
    generators : bool (default True)
        Only return the group generators if True.
    sympy_R: bool (default True)
        Whether the rotation matrices should be exact sympy
        representations.
    spin : float or sequence of arrays (optional)
        Spin representation to use for the unitary action of the
        operator. If not provided, the PointGroupElements have the unitary
        action set to None. If float is provided, it should be integer or half-integer
        specifying the spin representation in the standard basis, see `spin_matrices`.
        Otherwise a sequence of 3 arrays of identical square size must be provided
        representing 3 components of the angular momentum operator. The unitary action
        of rotation operator is `U = exp(-i n⋅s)`. In 2D the z axis is assumed to be
        the rotation axis. If `ph` is True, `spin` may not be provided, as it is not
        possible to deduce the unitary representation of particle-hole symmetry from
        spin alone. In this case construct the particle-hole operator manually.

    Returns:
    --------
    set of PointGroupElements

    Notes:
    ------
        Warning: in 2D the real space action of a mirror and and a 2-fold rotation
        around an axis in the plane is identical, however the action on angular momentum
        is different. Here we consider the action of the mirror, which is the same as the
        action of a 2-fold rotation around the mirror axis, assuming inversion acts trivially.
    """
    if spin is not None and sympy_R:
        U6 = spin_rotation(np.pi / 3 * np.array([0, 0, 1]), spin)
    else:
        U6 = None
    if dim == 2:
        Mx = mirror([1, 0], spin=spin)
        if sympy_R:

            C6 = PointGroupElement(sympy.ImmutableMatrix(
                                        [[sympy.Rational(1, 2), sympy.sqrt(3)/2],
                                         [-sympy.sqrt(3)/2,       sympy.Rational(1, 2)]]
                                                         ),
                                         False, False, U6)
        else:
            C6 = rotation(1/6, spin=spin)
        gens = {Mx, C6}
    elif dim == 3:
        I = inversion(3, U=(None if spin is None else spin_rotation(np.zeros(3), spin)))
        C2x = rotation(1/2, [1, 0, 0], spin=spin)
        if sympy_R:
            C6 = PointGroupElement(sympy.ImmutableMatrix(
                                        [[sympy.Rational(1, 2), sympy.sqrt(3)/2, 0],
                                         [-sympy.sqrt(3)/2, sympy.Rational(1, 2), 0],
                                         [0, 0, 1]]
                                                         ),
                                         False, False, U6)
        else:
            C6 = rotation(1/6, [0, 0, 1], spin=spin)
        gens = {I, C2x, C6}
    else:
        raise ValueError('Only 2 and 3 dimensions are supported.')

    if tr:
        TR = time_reversal(dim, spin=spin)
        gens.add(TR)
    if ph:
        PH = particle_hole(dim)
        gens.add(PH)
    if generators:
        return gens
    else:
        return generate_group(gens)


## Human readable group element names

def pretty_print_pge(g, full=False, latex=False):
    """
    Return a human readable string representation of PointGroupElement

    Parameters
    ----------

    g : PointGroupElement
        Point group element to be represented.
    full : bool (default False)
        Whether to return a full representation.
        The default short representation only contains the real space action
        and the symbol of the Altland-Zirnbauer part of the symmetry (see below).
        The full representation presents the symmetry action on the Hamiltonian
        and the unitary Hilbert-space action if set.
    latex : bool (default False)
        Whether to output LateX formatted string.

    Returns
    -------
    name : string
        In the short representation it is a sting `rot_name + az_name`.
        In the long representation the first line is the action on the
        Hamiltonian, the second line is `rot_name` and the third line
        is the unitary action as a matrix, if set.

        `rot_name` can be:
        - `1` for identity
        - `I` for inversion (in 1D mirror is the same as inversion)
        - `R(angle)` for 2D rotation
        - `R(angle, axis)` for 3D rotation (axis is not normalized)
        - `M(normal)` for mirror
        - `S(angle, axis)` for 3D rotoinversion (axis is not normalized)

        `az_name` can be:
        - `T` for time-reversal (antiunitary symmetry)
        - `P` for particle-hole (antiunitary antisymmetry)
        - `C` for chiral (unitary antisymmetry)
        - missing if the symmetry is unitary
    """

    def name_angle(theta, latex=False):
        frac = Fraction(theta / np.pi).limit_denominator(100)
        num, den = frac.numerator, frac.denominator
        if latex:
            if den == 1:
                angle = r'{}{}\pi'.format("-" if num < 0 else "",
                                         "" if abs(num) == 1 else abs(num))
            else:
                angle = r'{}\frac{{{}\pi}}{{{}}}'.format(
                            "-" if num < 0 else "",
                            "" if abs(num) == 1 else abs(num),
                            den)
        else:
            angle = '{}{}π{}'.format("-" if num < 0 else "",
                                     "" if abs(num) == 1 else abs(num),
                                     "" if den == 1 else ("/" + str(den)))
        return angle

    R = np.array(g.R).astype(float)
    if R.shape[0] == 1:
        if R[0, 0] == 1:
            rot_name = '1'
        else:
            rot_name = 'I'
    elif R.shape[0] == 2:
        if np.isclose(la.det(R), 1):
            # pure rotation
            theta = np.arctan2(R[1, 0], R[0, 0])
            if np.isclose(theta, 0):
                rot_name = '1'
            else:
                if latex:
                    rot_name = r'R\left({}\right)'.format(name_angle(theta, latex))
                else:
                    rot_name = 'R({})'.format(name_angle(theta))
        else:
            # mirror
            val, vec = la.eigh(R)
            assert allclose(val, [-1, 1]), R
            n = vec[:, 0]
            if latex:
                rot_name = r'M\left({}\right)'.format(_round_axis(n))
            else:
                rot_name = 'M({})'.format(_round_axis(n))
    elif R.shape[0] == 3:
        if np.isclose(la.det(R), 1):
            # pure rotation
            n, theta = rotation_to_angle(R)
            if np.isclose(theta, 0):
                rot_name = '1'
            else:
                if latex:
                    rot_name = (r'R\left({}, {}\right)'
                                .format(name_angle(theta, latex), _round_axis(n)))
                else:
                    rot_name = (r'R({}, {})'
                                .format(name_angle(theta, latex), _round_axis(n)))
        else:
            # rotoinversion
            n, theta = rotation_to_angle(-R)
            if np.isclose(theta, 0):
                # inversion
                rot_name = 'I'
            elif np.isclose(theta, np.pi):
                # mirror
                if latex:
                    rot_name = r'M\left({}\right)'.format(_round_axis(n))
                else:
                    rot_name = 'M({})'.format(_round_axis(n))
            else:
                # generic rotoinversion
                if latex:
                    rot_name = (r'S\left({}, {}\right)'
                                .format(name_angle(theta, latex), _round_axis(n)))
                else:
                    rot_name = ('S({}, {})'
                                .format(name_angle(theta, latex), _round_axis(n)))

    if full:
        if latex:
            name = r'U H(\mathbf{{k}}){} U^{{-1}} = {}H({}R\mathbf{{k}}) \\'
            name = name.format("^*" if g.conjugate else "", "-" if g.antisymmetry else "",
                               "-" if g.conjugate else "")
        else:
            name = '\nU⋅H(k){}⋅U^-1 = {}H({}R⋅k)\n'.format("*" if g.conjugate else "",
                                                         "-" if g.antisymmetry else "",
                                                         "-" if g.conjugate else "")
        name += 'R = {}'.format(rot_name) + (r'\\' if latex else '\n')
        if g.U is not None:
            if latex:
                Umat = _array_to_latex(np.round(g.U, 3))
                name += 'U = {}'.format(Umat)
            else:
                name += 'U = {}'.format(str(np.round(g.U, 3)).replace('\n', '\n    ')) +'\n\n'
    else:
        if g.conjugate and not g.antisymmetry:
            az_name = r" \mathcal{T}" if latex else "T"
        elif g.conjugate and g.antisymmetry:
            az_name = r" \mathcal{P}" if latex else "P"
        elif not g.conjugate and g.antisymmetry:
            az_name = r" \mathcal{C}" if latex else "C"
        else:
            az_name = ""
        name = (az_name if (rot_name == '1' and az_name != "")
                else rot_name + (" " if az_name is not "" else "") + az_name)
    return '$' + name + '$' if latex else name


def pretty_print_cgg(g, latex=False):
    """
    Return a human readable string representation of ContinuousGroupGenerator

    Parameters
    ----------

    g : ContinuousGroupGenerator
        Point group element to be represented.
    latex : bool (default False)
        Whether to output LateX formatted string.

    Returns
    -------
    name : string
        The first line is the action on the Hamiltonian, the following lines
        display the real space rotation as `R(ϕ)` for 2D rotation and
        `R(ϕ, axis)` for 3D rotation (axis is not normalized), and the conserved
        quantity as a matrix. If either of these is trivial, it is omitted.
        Note that a ContinuousGroupGenerator defines a continuous group
        of symmetries, so there is always a free parameter ϕ.
    """
    R, L = g.R, g.U
    # Find rotation axis
    if R is not None and allclose(R, np.zeros(R.shape)):
        R = None
    if L is not None and allclose(L, np.zeros(L.shape)):
        L = None

    if R is not None and R.shape[0] == 3:
        n = np.array([np.trace(l @ R) for l in L_matrices()]).real
        n /= la.norm(n)
        n = _round_axis(n)
        if latex:
            rot_name = r'R_{{\phi}} = R\left(\phi, {}\right)\\'.format(_round_axis(n))
        else:
            rot_name = '\nR_ϕ = R(ϕ, {})'.format(_round_axis(n))
    elif R is not None and R.shape[0] == 2:
        rot_name = r'R_{{\phi}} = R\left(\phi\right)\\' if latex else 'R_ϕ = R(ϕ)\n'
    else:
        rot_name = ''

    if L is not None:
        if latex:
            L_name = r'L = {} \\'.format(_array_to_latex(np.round(L, 3)))
        else:
            L_name = '\nL = {}'.format(str(np.round(L, 3)).replace('\n', '\n    '))

    if latex:
        if R is not None:
            name = r'{}H(\mathbf{{k}}){} = H(R_{{\phi}}\mathbf{{k}}) \\'
            name = name.format(r'e^{-i\phi L}' if L is not None else '',
                               r'e^{i\phi L}' if L is not None else '')
        else:
            name = r'\left[ H(\mathbf{{k}}), L \right] = 0 \\'
    else:
        if R is not None:
            name = '\n' + r'{}H(k){} = H(R_ϕ⋅k)'
            name = name.format(r'exp(-i ϕ L)⋅' if L is not None else '',
                               r'⋅exp(i ϕ L)' if L is not None else '')
        else:
            name = '\n[H(k), L] = 0'

    name += rot_name + L_name
    return '$' + name + '$' if latex else name


def rotation_to_angle(R):
    # Convert 3D rotation matrix to axis and angle
    assert allclose(R, R.real)
    n = np.array([1j * np.trace(L @ R) for L in L_matrices()]).real
    # Choose direction to minimize number of minus signs
    absn = la.norm(n) * np.sign(np.sum(n))
    if np.isclose(absn, 0):
        # n is zero for 2-fold rotation, find eigenvector with +1
        val, vec = la.eigh(R)
        assert np.isclose(val[-1], 1), R
        n = vec[:, -1]
        # Choose direction to minimize number of minus signs
        n /= np.sign(np.sum(n))
    else:
        n /= absn
    theta = np.arctan2(absn, np.trace(R).real - 1)
    return n, theta


def _round_axis(n):
    # Try to find integer axis
    for vec in product([-1, 0, 1], repeat=len(n)):
        if np.isclose(vec @ n, la.norm(vec)) and not np.isclose(la.norm(vec), 0):
            return np.array(vec, int)
    # otherwise round
    return np.round(n, 2)


def _array_to_latex(a, precision=2):
    """Returns a LaTeX bmatrix
    a: numpy array
    returns: LaTeX bmatrix as a string

    based on https://stackoverflow.com/questions/17129290/numpy-2d-and-1d-array-to-latex-bmatrix
    """
    if len(a.shape) > 2:
        raise ValueError('bmatrix can at most display two dimensions')
    rv = r'\begin{bmatrix}'
    for line in a:
        rv += np.array2string(line, precision=precision, separator='&', suppress_small=True)[1:-1] + r'\\'
    rv +=  r'\end{bmatrix}'
    return rv


## Predefined representations

def spin_matrices(s, include_0=False):
    """
    Construct spin-s matrices for any half-integer spin.

    Parameters:
    -----------

    s : float or int
        Spin representation to use, must be integer or half-integer.
    include_0 : bool (default False)
        If `include_0` is True, S[0] is the identity, indices 1, 2, 3
        correspond to x, y, z. Otherwise indices 0, 1, 2 are x, y, z.

    Returns:
    --------

    ndarray
        Sequence of spin-s operators in the standard spin-z basis.
        Array of shape `(3, 2*s + 1, 2*s + 1)`, or if `include_0` is True
        `(4, 2*s + 1, 2*s + 1)`.
    """
    d = np.round(2*s + 1)
    assert np.isclose(d, int(d))
    d = int(d)
    Sz = 1/2 * np.diag(np.arange(d - 1, -d, -2))
    # first diagonal for general s from en.wikipedia.org/wiki/Spin_(physics)
    diag = [1/2*np.sqrt((s + 1) * 2*i - i * (i + 1)) for i in np.arange(1, d)]
    Sx = np.diag(diag, k=1) + np.diag(diag, k=-1)
    Sy = -1j*np.diag(diag, k=1) + 1j*np.diag(diag, k=-1)
    if include_0:
        return np.array([np.eye(d), Sx, Sy, Sz])
    else:
        return np.array([Sx, Sy, Sz])


def spin_rotation(n, s, roundint=False):
    """
    Construct the unitary spin rotation matrix for rotation specified by the
    vector n (in radian units) with angular momentum `s`, given by
    `U = exp(-i n⋅s)`.

    Parameters:
    -----------

    n : iterable
        Rotation vector. Its norm is the rotation angle in radians.
    s : float or sequence of arrays
        Spin representation to use for the unitary action of the
        operator. If float is provided, it should be integer or half-integer
        specifying the spin representation in the standard basis, see `spin_matrices`.
        Otherwise a sequence of 3 arrays of identical square size must be provided
        representing 3 components of the angular momentum operator.
    roundint : bool (default False)
        If roundint is True, result is converted to integer tinyarray if possible.

    Returns:
    --------
    U : ndarray
        Unitary spin rotation matrix of the same shape as the spin matrices or
        `(2*s + 1, 2*s + 1)`.
    """
    n = np.array(n)
    if isinstance(s, Number):
        s = spin_matrices(s)
    else:
        s = np.array(s)
    # Make matrix exponential for rotation representation
    U = la.expm(-1j * np.tensordot(n, s, axes=((0), (0))))
    if roundint:
        Ur = np.round(np.real(U))
        assert allclose(U, Ur)
        U = ta.array(Ur.astype(int))
    return U


def L_matrices(d=3, l=1):
    """Construct real space rotation generator matrices in d=2 or 3 dimensions.
    Can also be used to get angular momentum operators for real atomic orbitals
    in 3 dimensions, for p-orbitals use `l=1`, for d-orbitals `l=2`. The basis
    of p-orbitals is `p_x`, `p_y`, `p_z`, for d-orbitals `d_{x^2 - y^2}`,
    `d_{3 z^2 - r^2}`, `d_{xy}`, `d_{yz}`, `d_{zx}`. The matrices are all
    purely imaginary and antisymmetric.
    To generate finite rotations, use 'spin_rotation(n, L)'.
    """
    if d == 2 and l==1:
        return 1j * np.array([[[0, -1],
                              [1, 0]]])
    elif d == 3 and l==1:
        return 1j * np.array([[[0, 0, 0],
                               [0, 0, -1],
                               [0, 1, 0]],
                              [[0, 0, 1],
                               [0, 0, 0],
                               [-1, 0, 0]],
                              [[0, -1, 0],
                               [1, 0, 0],
                               [0, 0, 0]]])
    elif d == 3 and l==2:
        s3 = np.sqrt(3)
        return 1j * np.array([[[0, 0, 0, -1, 0],
                              [0, 0, 0, -s3, 0],
                              [0, 0, 0, 0, 1],
                              [1, s3, 0, 0, 0],
                              [0, 0, -1, 0, 0]],
                             [[0, 0, 0, 0, -1],
                              [0, 0, 0, 0, s3],
                              [0, 0, 0, -1, 0],
                              [0, 0, 1, 0, 0],
                              [1, -s3, 0, 0, 0]],
                             [[0, 0, -2, 0, 0],
                              [0, 0, 0, 0, 0],
                              [2, 0, 0, 0, 0],
                              [0, 0, 0, 0, 1],
                              [0, 0, 0, -1, 0]]])
    else:
        raise ValueError('Only 2 and 3 dimensions are supported.')

## Utility function for lattice models

def symmetry_from_permutation(R, perm, norbs, onsite_action=None,
                              antiunitary=False, antisymmetry=False):
    """Construct symmetry operator for lattice systems with multiple sites.

    Parameters:
    -----------
    R : real space rotation
    perm : dict : {site: image_site}
        permutation of the sites under the symmetry action
    norbs : OrderedDict : {site : norbs_site} or tuple of tuples ((site, norbs_site), )
        sites are ordered in the order specified, with blocks of size norbs_site
        corresponding to each site.
    onsite_action : dict : {site: ndarray} or ndarray or None
        onsite symmetry action, such as spin rotation for each site. If only one
        array is specified, it is used for every site. If None (default), identity
        is used on every site. Size of the arrays must be consistent with norbs.
    antiunitary, antisymmetry : bool

    Returns:
    --------
    g : PointGroupElement
        PointGroupElement corresponding to the operation.

    Notes:
    ------
    Sites can be indexed by any hashable identifiers, such as integers or stings.
    """
    norbs = OrderedDict(norbs)
    ranges = dict()
    N = 0
    for a, n in norbs.items():
        ranges[a] = slice(N, N + n)
        N += n

    if onsite_action is None:
        onsite_action = {a: np.eye(n) for a, n in norbs.items()}
    elif isinstance(onsite_action, np.ndarray):
        onsite_action = {a: onsite_action for a in norbs.keys()}

    # Build transformation matrix
    if not set(perm.keys()) == set(perm.values()) == set(norbs.keys()):
        raise ValueError('perm keys, values and norbs keys must contain the same sites.')
    U = np.zeros((N, N), dtype=complex)
    for a in norbs.keys():
        if not norbs[a] == norbs[perm[a]]:
            raise ValueError('Symmetry related sites must have the same number or orbitals')
        U[ranges[a], ranges[perm[a]]] = onsite_action[a]
    g = PointGroupElement(R, antiunitary, antisymmetry, U)
    return g


class PrettyList(list):
    """Subclass of list that displays its elements with latex."""
    def _repr_latex_(self):
        return (
            r'$'
            + r'\\'.join(i._repr_latex_().replace('$', '') for i in self)
            + r'$'
        )



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

# =============
# Kwant license
# =============
#
# Copyright 2011-2015 C. W. Groth (CEA), M. Wimmer, A. R. Akhmerov,
# X. Waintal (CEA), and others.  All rights reserved.
#
# (CEA = Commissariat à l'énergie atomique et aux énergies alternatives)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
# ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
# ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

__all__ = ['lll', 'cvp', 'voronoi']

import numpy as np
from scipy import linalg as la
from itertools import product


def gs_coefficient(a, b):
    """Gram-Schmidt coefficient."""
    return np.dot(a, b) / np.linalg.norm(b)**2


def gs(mat):
    """Compute Gram-Schmidt decomposition on a matrix."""
    mat = np.copy(mat)
    for i in range(len(mat)):
        for j in range(i):
            mat[i] -= gs_coefficient(mat[i], mat[j]) * mat[j]
    return mat


def is_c_reduced(vecs, c):
    """Check if a basis is c-reduced."""
    vecs = gs(vecs)
    r = np.apply_along_axis(lambda x: np.linalg.norm(x)**2, 1, vecs)
    return np.all((r[: -1] / r[1:]) < c)


def lll(basis, c=1.34):
    """
    Calculate a reduced lattice basis using LLL algorithm.

    Reduce a basis of a lattice to an almost orthonormal form. For details see
    e.g. http://en.wikipedia.org/wiki/LLL-algorithm.

    Parameters
    ----------
    basis : 2d array-like of floats
        The lattice basis vectors to be reduced.
    c : float
        Reduction parameter for the algorithm. Must be larger than 1 1/3,
        since otherwise a solution is not guaranteed to exist.

    Returns
    -------
    reduced_basis : numpy array
        The basis vectors of the LLL-reduced basis.
    transformation : numpy integer array
        Coefficient matrix for tranforming from the reduced basis to the
        original one.
    """
    vecs_orig = np.asarray(basis, float)
    if vecs_orig.ndim != 2:
        raise ValueError('"basis" must be a 2d array-like object.')
    if vecs_orig.shape[0] > vecs_orig.shape[1]:
        raise ValueError('The number of basis vectors exceeds the '
                         'space dimensionality.')
    vecs = vecs_orig.copy()
    vecsstar = vecs_orig.copy()
    m = vecs.shape[0]
    u = np.identity(m)
    def ll_reduce(i):
        for j in reversed(range(i)):
            vecs[i] -= np.round(u[i, j]) * vecs[j]
            u[i] -= np.round(u[i, j]) * u[j]

    # Initialize values.
    for i in range(m):
        for j in range(i):
            u[i, j] = gs_coefficient(vecs[i], vecsstar[j])
            vecsstar[i] -= u[i, j] * vecsstar[j]
        ll_reduce(i)

    # Main part of LLL algorithm.
    i = 0
    while i < m-1:
        if (np.linalg.norm(vecsstar[i]) ** 2 <
           c * np.linalg.norm(vecsstar[i+1]) ** 2):
            i += 1
        else:
            vecsstar[i+1] += u[i+1, i] * vecsstar[i]

            u[i, i] = gs_coefficient(vecs[i], vecsstar[i+1])
            u[i, i+1] = u[i+1, i] = 1
            u[i+1, i+1] = 0
            vecsstar[i] -= u[i, i] * vecsstar[i+1]
            vecs[[i, i+1]] = vecs[[i+1, i]]
            vecsstar[[i, i+1]] = vecsstar[[i+1, i]]
            u[[i, i+1]] = u[[i+1, i]]
            for j in range(i+2, m):
                u[j, i] = gs_coefficient(vecs[j], vecsstar[i])
                u[j, i+1] = gs_coefficient(vecs[j], vecsstar[i+1])
            if abs(u[i+1, i]) > 0.5:
                ll_reduce(i+1)
            i = max(i-1, 0)
    # TODO: change to rcond=None once we depend on numpy >= 1.14.
    coefs = np.linalg.lstsq(vecs_orig.T, vecs.T, rcond=-1)[0]
    if not np.allclose(np.round(coefs), coefs, atol=1e-6):
        raise RuntimeError('LLL algorithm instability.')
    if not is_c_reduced(vecs, c):
        raise RuntimeError('LLL algorithm instability.')
    return vecs, np.array(np.round(coefs), int)


def cvp(vec, basis, n=1, group_by_length=False, rtol=1e-09):
    """
    Solve the n-closest vector problem for a vector, given a basis.

    This algorithm performs poorly in general, so it should be supplied
    with LLL-reduced basis.

    Parameters
    ----------
    vec : 1d array-like of floats
        The lattice vectors closest to this vector are to be found.
    basis : 2d array-like of floats
        Sequence of basis vectors
    n : int
        Number of lattice vectors closest to the point that need to be found. If
        `group_by_length=True`, all vectors with the `n` shortest distinct distances
        are found.
    group_by_length : bool
        Whether to count points with distances that are within `rtol` as one
        towards `n`. Useful for finding `n`-th nearest neighbors of high symmetry points.
    rtol : float
        If `group_by_length=True`, relative tolerance when deciding whether
        distances are equal.

    Returns
    -------
    coords : numpy array
        An array with the coefficients of the `n` closest lattice vectors, or all
        the lattice vectors with the `n` shortest distances from the
        requested point.

    Notes
    -----
    This function can also be used to solve the `n` shortest lattice vector
    problem if the `vec` is zero, and `n+1` points are requested
    (and the first output is ignored).
    """
    # Calculate coordinates of the starting point in this basis.
    basis = np.asarray(basis)
    if basis.ndim != 2:
        raise ValueError('`basis` must be a 2d array-like object.')
    vec = np.asarray(vec)
    # Project the coordinates on the space spanned by `basis`, in
    # units of `basis` vectors.
    vec_bas = la.lstsq(basis.T, vec)[0]
    # Round the coordinates, this finds the lattice point which is
    # the center of the parallelepiped that contains `vec`.
    # `vec` is surely in the parallelepiped with edges `basis`
    # centered at `center_coords`.
    center_coords = np.array(np.round(vec_bas), int)
    # Make `vec` the projection of `vec` onto the space spanned by `basis`.
    vec = vec_bas @ basis
    bbt = basis @ basis.T
    # The volume of a parallelepiped spanned by `basis` is `sqrt(det(bbt))`.
    # (We use this instead of `det(basis)` because basis does not necessarily
    # span the full space, hence not always a square matrix.)
    # We calculate the radius `rad` of the largest sphere that can be inscribed
    # in the parallelepiped spanned by `basis`. The area of each face spanned
    # by one less vector in `basis` is given by the above formula applied to `bbt`
    # with one row and column discarded (first minor). We utilizie the relation
    # between entries of the inverse matrix and first minors. The perpendicular
    # size of the parallelepiped is given by the volume of the parallelepiped
    # divided by the area of the face. We choose the smallest perpendicular
    # size as the diameter of the largest sphere inside.
    rad = 0.5 / np.sqrt(np.max(np.diag(la.inv(bbt))))

    l = 1
    while True:
        # Make all the lattice points in and on the edges of a parallelepiped
        # of `2*l` size around `center_coords`.
        points = np.mgrid[tuple(slice(i - l, i + l + 1) for i in center_coords)]
        points = points.reshape(basis.shape[0], -1).T
        # If there are less than `n` points, make more.
        if len(points) < n:
            l += 1
            continue
        point_coords = points @ basis
        point_coords = point_coords - vec.T
        distances = la.norm(point_coords, axis = 1)
        order = np.argsort(distances)
        distances = distances[order]
        points = points[order]
        if group_by_length:
            # Group points that are equidistant within `rtol * rad`.
            group_boundaries = np.where(np.diff(distances) > rtol * rad)[0]
            if len(group_boundaries) + 1 < n:
                # Make more points if there are less than `n` groups.
                l += 1
                continue
            elif len(group_boundaries) == n - 1:
                # If there are exactly `n` groups, choose largest distance.
                dist_n = distances[-1]
                points_keep = points
            else:
                # If there are more than `n` groups, choose the largest
                # distance in the `n`-th group.
                dist_n = distances[group_boundaries[n-1]]
                # Only return points that are in the first `n` groups.
                points_keep = points[:group_boundaries[n-1] + 1]
        else:
            rtol = 0
            # Choose the `n`-th distance.
            dist_n = distances[n-1]
            # Only return the first `n` points.
            points_keep = points[:n]
        # We covered all points within radius `(2*l-1) * rad` from the
        # parallelepiped spanned by `basis` centered at `center_coords`.
        # Because `vec` is guaranteed to be in this parallelepiped, if
        # the current `dist_n` is smaller than `(2*l-1) * rad`, we surely
        # found all the smaller vectors.
        if dist_n < (2*l - 1 - rtol) * rad:
            return np.array(points_keep, int)
        else:
            # Otherwise there may be smaller vectors we haven't found,
            # increase `l`.
            l += 1
            continue


def voronoi(basis, reduced=False, rtol=1e-09):
    """
    Return an array of lattice vectors forming its voronoi cell.

    Parameters
    ----------
    basis : 2d array-like of floats
        Basis vectors for which the Voronoi neighbors have to be found.
    reduced : bool
        If False, exactly `2 (2^D - 1)` vectors are returned (where `D`
        is the number of lattice vectors), these are not always the minimal
        set of Voronoi vectors.
        If True, only the minimal set of Voronoi vectors are returned.
    rtol : float
        Tolerance when deciding whether a vector is in the minimal set,
        vectors associated with small faces compared to the size of the
        unit cell are discarded.

    Returns
    -------
    voronoi_neighbors : numpy array of ints
        All the lattice vectors that (potentially) neighbor the origin.
        List of length `2n` where the second half of the list contains
        is `-1` times the vectors in the first half.
    """
    basis = np.asarray(basis)
    if basis.ndim != 2:
        raise ValueError('`basis` must be a 2d array-like object.')
    # Find halved lattice points, every face of the VC contains half
    # of the lattice vector which is the normal of the face.
    # These points are all potentially on a face a VC,
    # but not necessarily the VC centered at the origin.
    displacements = list(product(*(len(basis) * [[0, .5]])))[1:]
    # Find the nearest lattice point, this is the lattice point whose
    # VC this face belongs to.
    vertices = np.array([cvp(vec @ basis, basis)[0] for vec in
                         displacements])
    # The lattice vector for a face is exactly twice the vector to the
    # closest lattice point from the halved lattice point on the face.
    vertices = np.array(np.round((vertices - displacements) * 2), int)
    if reduced:
        # Discard vertices that are not associated with a face of the VC.
        # This happens if half of the lattice vector is outside the convex
        # polytope defined by the rest of the vertices in `keep`.
        bbt = basis @ basis.T
        products = vertices @ bbt @ vertices.T
        keep = np.array([True] * len(vertices))
        for i in range(len(vertices)):
            # Relevant if the projection of half of the lattice vector `vertices[i]`
            # onto every other lattice vector in `veritces[keep]` is less than `0.5`.
            mask = np.array([False if k == i else b for k, b in enumerate(keep)])
            projections = 0.5 * products[i, mask] / np.diag(products)[mask]
            if not np.all(np.abs(projections) < 0.5 - rtol):
                keep[i] = False
        vertices = vertices[keep]

    vertices = np.concatenate([vertices, -vertices])
    return vertices
